# Tests of the alloc service

import unittest

import calipertest as cat

class CaliperAllocServiceTest(unittest.TestCase):
    """ Caliper alloc service test cases """

    def test_alloc(self):
        target_cmd = [ './ci_test_alloc' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'        : 'alloc:recorder:trace:event',
            'CALI_EVENT_ENABLE_SNAPSHOT_INFO' : 'false',
            'CALI_ALLOC_RESOLVE_ADDRESSES': 'true',
            'CALI_RECORDER_FILENAME'      : 'stdout',
            'CALI_LOG_VERBOSITY'          : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        # test allocated.0

        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, { 'mem.alloc', 'alloc.uid', 'alloc.address', 'alloc.total_size' }))
        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, { 'mem.free', 'alloc.uid', 'alloc.address', 'alloc.total_size' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'mem.alloc': 'test_alloc_A' }))
        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, { 'mem.free': 'test_alloc_A'  }))

        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, { 'test_alloc.allocated.0', 'ptr_in', 'ptr_out' }))

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, {'test_alloc.allocated.0' : 'true',
                        'alloc.uid#ptr_in'       : '1',
                        'alloc.index#ptr_in'     : '0',
                        'alloc.label#ptr_in'     : 'test_alloc_A'
                    }))

        self.assertFalse(cat.has_snapshot_with_keys(
            snapshots, { 'test_alloc.allocated.0', 'alloc.uid#ptr_out' }))

        # test allocated.1

        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, { 'test_alloc.allocated.1', 'ptr_in', 'ptr_out' }))

        self.assertTrue(cat.has_snapshot_with_attributes(
            snapshots, {'test_alloc.allocated.1' : 'true',
                        'alloc.uid#ptr_in'       : '1',
                        'alloc.index#ptr_in'     : '41',
                        'alloc.label#ptr_in'     : 'test_alloc_A'
                    }))

        self.assertFalse(cat.has_snapshot_with_keys(
            snapshots, { 'test_alloc.allocated.1', 'alloc.uid#ptr_out' }))

        # test allocated.freed

        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, { 'test_alloc.freed', 'ptr_in', 'ptr_out' }))

        self.assertFalse(cat.has_snapshot_with_keys(
            snapshots, { 'test_alloc.freed', 'alloc.uid#ptr_in' }))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import hatchet as ht


if __name__ == "__main__":
    # Path to caliper cali file.
    cali_file = (
        "../../../hatchet/tests/data/caliper-lulesh-cali/lulesh-annotation-profile.cali"
    )

    # Setup desired cali query.
    grouping_attribute = "function"
    default_metric = "sum(sum#time.duration),inclusive_sum(sum#time.duration)"
    query = "select function,%s group by %s format json-split" % (
        default_metric,
        grouping_attribute,
    )

    # Use hatchet's ``from_caliper`` API with the path to the cali file and the
    # query. This API will internally run ``cali-query`` on this file to
    # produce a json-split stream. The result is stored into Hatchet's
    # GraphFrame.
    gf = ht.GraphFrame.from_caliper(cali_file, query)

    # Printout the DataFrame component of the GraphFrame.
    print(gf.dataframe)

    # Printout the graph component of the GraphFrame.
    # Use "time (inc)" as the metric column to be displayed
    print(gf.tree(metric_column="time (inc)"))



#!/usr/bin/env python

import hatchet as ht


if __name__ == "__main__":
    # Path to caliper json-split file.
    json_file = "../../../hatchet/tests/data/caliper-cpi-json/cpi-callpath-profile.json"

    # Use hatchet's ``from_caliper_json`` API with the resulting json-split.
    # The result is stored into Hatchet's GraphFrame.
    gf = ht.GraphFrame.from_caliper_json(json_file)

    # Printout the DataFrame component of the GraphFrame.
    print(gf.dataframe)

    # Printout the graph component of the GraphFrame.
    # Because no metric parameter is specified, ``time`` is used by default.
    print(gf.tree())



# MIT License
#
# Copyright (c) 2020, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" @file plotting.py
Table generation routines for timemory package
"""

from __future__ import absolute_import
from __future__ import division

__author__ = "Jonathan Madsen"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Jonathan Madsen"]
__license__ = "MIT"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"
__all__ = [
    "timem",
]

import os
import sys
from .plotting import make_output_directory, add_plotted_files

# use the matplotlib import stuff from plotting
try:
    import matplotlib.pyplot as plt

except ImportError as e:
    print(f"{e}")
    pass


verbosity = os.environ.get("TIMEMORY_VERBOSE", 0)


def timem(data, fname, args):
    cmd = "unknown"
    if "command_line" in data:
        cmd = data["command_line"][1]

    nranks = len(data["timem"])
    row_labels = []
    col_labels = []
    table_vals = []

    min_data = {}
    max_data = {}
    sum_data = {}
    int_data = {}

    for rank in range(nranks):
        col_labels.append("Rank {}".format(rank))
        i = 0
        for key, entry in data["timem"][rank].items():
            if isinstance(entry["value"], (list, dict)):
                sys.stderr.write(
                    f"Skipping multi-dimensional entry for {key}\n"
                )
                continue
            if rank == 0:
                unit = entry["unit_repr"]
                if unit:
                    unit = " [{}]".format(unit)
                lbl = "{}{}".format(key.replace("_", " ").title(), unit)
                row_labels.append(lbl)
                min_data[i] = entry["repr"]
                max_data[i] = entry["repr"]
                sum_data[i] = 0
            if len(table_vals) < i + 1:
                table_vals.append([])
            value = entry["repr"]
            strval = "{}".format(value)
            min_data[i] = min([min_data[i], value])
            max_data[i] = max([max_data[i], value])
            sum_data[i] += value
            if "." in strval:
                table_vals[i].append("{:.3f}".format(value))
                int_data[i] = False
            else:
                table_vals[i].append("{}".format(value))
                if i not in int_data:
                    int_data[i] = True
            i += 1

    if nranks > 1:
        col_labels = ["Sum", "Mean"] + col_labels + ["Min", "Max"]
        i = 0
        for i in range(len(row_labels)):
            _mean = sum_data[i] / nranks
            if int_data[i]:
                table_vals[i] = (
                    ["{}".format(sum_data[i]), "{:.3f}".format(_mean)]
                    + table_vals[i]
                    + ["{}".format(min_data[i]), "{}".format(max_data[i])]
                )
            else:
                table_vals[i] = (
                    ["{:.3f}".format(sum_data[i]), "{:.3f}".format(_mean)]
                    + table_vals[i]
                    + [
                        "{:.3f}".format(min_data[i]),
                        "{:.3f}".format(max_data[i]),
                    ]
                )
            i += 1

    # print(f"cols: {col_labels}")
    # print(f"rows: {row_labels}")
    # print(f"vals: {table_vals}")
    #
    font = {
        "family": "serif",
        "color": "black",
        "weight": "bold",
        "size": 20,
    }

    if args.titles is not None:
        plt.suptitle("{}".format(args.titles[0].title()), **font)
    else:
        plt.suptitle(f"{cmd}", **font)

    # Draw table
    the_table = plt.table(
        cellText=table_vals,
        colWidths=[0.2] * len(col_labels),
        rowLabels=row_labels,
        colLabels=col_labels,
        loc="center",
    )
    the_table.auto_set_font_size(False)
    the_table.set_fontsize(16)
    the_table.scale(2, 2)

    # Removing ticks and spines enables you to get the figure only with table
    plt.tick_params(
        axis="x", which="both", bottom=False, top=False, labelbottom=False
    )
    plt.tick_params(
        axis="y", which="both", right=False, left=False, labelleft=False
    )
    for pos in ["right", "top", "bottom", "left"]:
        plt.gca().spines[pos].set_visible(False)

    output_dir = "."
    try:
        output_dir = args.output_dir
    except AttributeError:
        pass

    if args.display_plot:
        print("Displaying plot...")
        plt.show()
    else:
        make_output_directory(output_dir)
        imgfname = (
            os.path.basename(fname)
            .replace(".json", ".{}".format(args.img_type))
            .replace(".py", ".{}".format(args.img_type))
        )
        if not ".{}".format(args.img_type) in imgfname:
            imgfname += ".{}".format(args.img_type)
        while ".." in imgfname:
            imgfname = imgfname.replace("..", ".")
        add_plotted_files(
            imgfname, os.path.join(output_dir, imgfname), args.echo_dart
        )

        imgfname = os.path.join(output_dir, imgfname)
        if verbosity > 0:
            print("Opening '{}' for output...".format(imgfname))
        else:
            lbl = os.path.basename(imgfname).strip(".{}".format(args.img_type))
            print("[{}]|0> Outputting '{}'...".format(lbl, imgfname))
        plt.savefig(imgfname, bbox_inches="tight", pad_inches=0.05)
        plt.close()
        if verbosity > 0:
            print("Closed '{}'...".format(imgfname))

    # fig, ax = plt.subplots(figsize=(16, 8))

    # for step, group in df.groupby('step', observed=True):
    #     d = group['ipc'].rolling('30s').agg(['mean', 'std']).resample('10s').mean()
    #     mean = d['mean']
    #     std = d['std']
    #     mean.plot(ax=ax, label=step, color=color)
    #     std = d['std']
    #     ax.fill_between(std.index, mean - 0.5*std, mean + 0.5*std, alpha=0.2,
    #                     color=color)

    # _ = ax.set_title(f'Instructions Per Cycle (IPC) by step jobid = {jobid}')
    # _ = ax.set_ylabel('IPC')
    # _ = ax.legend(ncol=2, title='job step')

    # buf = io.BytesIO()
    # plt.savefig(buf,  format='png')
    # buf.seek(0)
    # return base64.b64encode(buf.read())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from os.path import dirname, join, abspath
import sys
import os

if __name__ == '__main__':
    cwd = os.getcwd()
    repo_dir = abspath(dirname(__file__))
    test_dir = join(repo_dir, 'tests')
    print('cwd = {!r}'.format(cwd))

    import pytest

    # Prefer testing the installed version, but fallback to testing the
    # development version.
    try:
        import ubelt as ub
    except ImportError:
        print('running this test script requires ubelt')
        raise
    # Statically check if ``line_profiler`` is installed outside of the repo.
    # To do this, we make a copy of PYTHONPATH, remove the repodir, and use
    # ubelt to check to see if ``line_profiler`` can be resolved to a path.
    temp_path = list(map(abspath, sys.path))
    if repo_dir in temp_path:
        temp_path.remove(repo_dir)
    modpath = ub.modname_to_modpath('line_profiler', sys_path=temp_path)
    if modpath is not None:
        # If it does, then import it. This should cause the installed version
        # to be used on further imports even if the repo_dir is in the path.
        print('Using installed version of line_profiler')
        module = ub.import_module_from_path(modpath, index=0)
        print('Installed module = {!r}'.format(module))
    else:
        print('No installed version of line_profiler found')

    try:
        print('Changing dirs to test_dir={!r}'.format(test_dir))
        os.chdir(test_dir)

        package_name = 'line_profiler'
        pytest_args = [
            '--cov-config', '../.coveragerc',
            '--cov-report', 'html',
            '--cov-report', 'term',
            '--cov=' + package_name,
        ]
        pytest_args = pytest_args + sys.argv[1:]
        sys.exit(pytest.main(pytest_args))
    finally:
        os.chdir(cwd)
        print('Restoring cwd = {!r}'.format(cwd))



# Callpath sample report tests

import json
import os
import unittest

import calipertest as cat

class CaliperSampleReportTest(unittest.TestCase):
    """ Callpath sample report controller """
    def test_sample_report_nompi(self):
        target_cmd = [ './ci_test_macros', '5000', 'callpath-sample-report,output=stdout' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'      : '0',
        }

        log_targets = [
            'Path',
            'main',
            '  foo'
        ]

        report_out,_ = cat.run_test(target_cmd, caliper_config)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

    def test_runtime_report_nompi(self):
        target_cmd = [ './ci_test_macros', '5000', 'callpath-sample-report,aggregate_across_ranks=false,output=stdout' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'      : '0',
        }

        log_targets = [
            'Samples Time (sec)',
            'main',
            '  foo'
        ]

        report_out,_ = cat.run_test(target_cmd, caliper_config)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

    def test_hatchet_callpath_sample_profile(self):
        target_cmd = [ './ci_test_macros', '5000', 'hatchet-sample-profile(use.mpi=false,output=stdout,output.format=cali)' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) > 0)

        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, { 'loop', 'function', 'source.function#callpath.address' }))

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from os.path import exists
import os
import sys
import setuptools  # NOQA
from setuptools import find_packages


def parse_version(fpath):
    """
    Statically parse the version number from a python file
    """
    import ast
    if not exists(fpath):
        raise ValueError('fpath={!r} does not exist'.format(fpath))
    with open(fpath, 'r') as file_:
        sourcecode = file_.read()
    pt = ast.parse(sourcecode)
    class VersionVisitor(ast.NodeVisitor):
        def visit_Assign(self, node):
            for target in node.targets:
                if getattr(target, 'id', None) == '__version__':
                    self.version = node.value.s
    visitor = VersionVisitor()
    visitor.visit(pt)
    return visitor.version


def parse_description():
    """
    Parse the description in the README file

    CommandLine:
        pandoc --from=markdown --to=rst --output=README.rst README.md
        python -c "import setup; print(setup.parse_description())"
    """
    from os.path import dirname, join, exists
    readme_fpath = join(dirname(__file__), 'README.rst')
    # This breaks on pip install, so check that it exists.
    if exists(readme_fpath):
        with open(readme_fpath, 'r') as f:
            text = f.read()
        return text
    return ''


def parse_requirements(fname='requirements.txt', with_version=False):
    """
    Parse the package dependencies listed in a requirements file but strips
    specific versioning information.

    Args:
        fname (str): path to requirements file
        with_version (bool, default=False): if true include version specs

    Returns:
        List[str]: list of requirements items

    CommandLine:
        python -c "import setup; print(setup.parse_requirements())"
        python -c "import setup; print(chr(10).join(setup.parse_requirements(with_version=True)))"
    """
    from os.path import exists
    import re
    require_fpath = fname

    def parse_line(line):
        """
        Parse information from a line in a requirements text file
        """
        if line.startswith('-r '):
            # Allow specifying requirements in other files
            target = line.split(' ')[1]
            for info in parse_require_file(target):
                yield info
        else:
            info = {'line': line}
            if line.startswith('-e '):
                info['package'] = line.split('#egg=')[1]
            else:
                # Remove versioning from the package
                pat = '(' + '|'.join(['>=', '==', '>']) + ')'
                parts = re.split(pat, line, maxsplit=1)
                parts = [p.strip() for p in parts]

                info['package'] = parts[0]
                if len(parts) > 1:
                    op, rest = parts[1:]
                    if ';' in rest:
                        # Handle platform specific dependencies
                        # http://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-platform-specific-dependencies
                        version, platform_deps = map(str.strip, rest.split(';'))
                        info['platform_deps'] = platform_deps
                    else:
                        version = rest  # NOQA
                    info['version'] = (op, version)
            yield info

    def parse_require_file(fpath):
        with open(fpath, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line and not line.startswith('#'):
                    for info in parse_line(line):
                        yield info

    def gen_packages_items():
        if exists(require_fpath):
            for info in parse_require_file(require_fpath):
                parts = [info['package']]
                if with_version and 'version' in info:
                    parts.extend(info['version'])
                if not sys.version.startswith('3.4'):
                    # apparently package_deps are broken in 3.4
                    platform_deps = info.get('platform_deps')
                    if platform_deps is not None:
                        parts.append(';' + platform_deps)
                item = ''.join(parts)
                yield item

    packages = list(gen_packages_items())
    return packages


def native_mb_python_tag():
    import sys
    import platform
    major = sys.version_info[0]
    minor = sys.version_info[1]
    ver = '{}{}'.format(major, minor)
    if platform.python_implementation() == 'CPython':
        # TODO: get if cp27m or cp27mu
        impl = 'cp'
        if ver == '27':
            IS_27_BUILT_WITH_UNICODE = True  # how to determine this?
            if IS_27_BUILT_WITH_UNICODE:
                abi = 'mu'
            else:
                abi = 'm'
        else:
            if ver == '38':
                # no abi in 38?
                abi = ''
            else:
                abi = 'm'
    else:
        raise NotImplementedError(impl)
    mb_tag = '{impl}{ver}-{impl}{ver}{abi}'.format(**locals())
    return mb_tag


USE_SKBUILD = True
if USE_SKBUILD:

    if '--universal' in sys.argv:
        # Dont use scikit-build for universal wheels
        if 'develop' in sys.argv:
            sys.argv.remove('--universal')
        from setuptools import setup
    else:
        from skbuild import setup

    setupkw = dict(
        # include_package_data=False,
        # package_dir={
        #     '': '.',
        #     # Note: this requires that FLANN_LIB_INSTALL_DIR is set to pyflann/lib
        #     # in the src/cpp/CMakeLists.txt
        #     # 'line_profiler.lib': 'line_profiler/lib',
        # },
    )
else:
    # Monkeypatch distutils.
    import distutils.errors
    # from distutils.core import setup
    from distutils.extension import Extension
    from distutils.log import warn
    from setuptools import setup
    # use setuptools
    try:
        from Cython.Distutils import build_ext
        cmdclass = dict(build_ext=build_ext)
        line_profiler_source = '_line_profiler.pyx'
    except ImportError:
        cmdclass = {}
        line_profiler_source = '_line_profiler.c'
        if not os.path.exists(line_profiler_source):
            raise distutils.errors.DistutilsError("""\
    You need Cython to build the line_profiler from a git checkout, or
    alternatively use a release tarball from PyPI to build it without Cython.""")
        else:
            warn("Could not import Cython. "
                 "Using the available pre-generated C file.")

    setupkw = dict(
        cmdclass=cmdclass,
        ext_modules=[
            Extension('_line_profiler',
                      sources=[line_profiler_source, 'timers.c', 'unset_trace.c'],
                      depends=['python25.pxd']),
        ],
    )

long_description = """\
line_profiler will profile the time individual lines of code take to execute.
The profiler is implemented in C via Cython in order to reduce the overhead of
profiling.

Also included is the script kernprof.py which can be used to conveniently
profile Python applications and scripts either with line_profiler or with the
function-level profiling tools in the Python standard library.
"""

VERSION = parse_version('line_profiler/line_profiler.py')
MB_PYTHON_TAG = native_mb_python_tag()
# note: name is temporary until line_profiler pypi name is transfered
NAME = 'line_profiler'


# py_modules = ['line_profiler', 'kernprof']
# if sys.version_info > (3, 4):
#     py_modules += ['line_profiler_py35']


if __name__ == '__main__':
    setupkw.update(dict(
        name=NAME,
        version=VERSION,
        author='Robert Kern',
        author_email='robert.kern@enthought.com',
        description='Line-by-line profiler.',
        long_description=long_description,
        long_description_content_type='text/x-rst',
        url='https://github.com/pyutils/line_profiler',
        license="BSD",
        keywords=['timing', 'timer', 'profiling', 'profiler', 'line_profiler'],
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: BSD License",
            "Operating System :: OS Independent",
            "Programming Language :: C",
            "Programming Language :: Python",
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.2',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: Implementation :: CPython',
            "Topic :: Software Development",
        ],
        # py_modules=find_packages(),
        packages=list(find_packages()),
        py_modules=['kernprof', 'line_profiler'],
        entry_points={
            'console_scripts': [
                'kernprof=kernprof:main',
            ],
        },
        install_requires=parse_requirements('requirements/runtime.txt'),
        extras_require={
            'all': parse_requirements('requirements.txt'),
            'tests': parse_requirements('requirements/tests.txt'),
            'build': parse_requirements('requirements/build.txt'),
        },
    ))
    setup(**setupkw)



##############################################################################
# Copyright (c) 2017, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory.
# 
# This file is part of Caliper.
# Created by David Boehme, boehme3@llnl.gov, All rights reserved.
# LLNL-CODE-678900
# 
# For details, see https://github.com/scalability-llnl/Caliper.
# Please also see the LICENSE file for our additional BSD notice.
# 
# Redistribution and use in source and binary forms, with or without modification, are
# permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this list of
#   conditions and the disclaimer below.
# * Redistributions in binary form must reproduce the above copyright notice, this list of
#   conditions and the disclaimer (as noted below) in the documentation and/or other materials
#   provided with the distribution.
# * Neither the name of the LLNS/LLNL nor the names of its contributors may be used to endorse
#   or promote products derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
# LAWRENCE LIVERMORE NATIONAL SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##############################################################################

#
# Utilities and helper functions for the Caliper application/workflow test framework 
#

import os
import subprocess

class CaliperExecError(Exception):
    """ Error class when running a process fails
    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)
    
    
def run_test_with_query(target_cmd, query_cmd, env):
    """ Execute a command and query, and return query command's output
    
        This will run the command given by target_cmd with the environment given 
        by env, and pipe its stdout output into the command given by query_cmd.
    """

    target_proc = subprocess.Popen(target_cmd, env=env, stdout=subprocess.PIPE)
    query_proc  = subprocess.Popen(query_cmd, stdin=target_proc.stdout, stdout=subprocess.PIPE)
    target_proc.stdout.close()
    query_out, query_err = query_proc.communicate()
    target_proc.wait()

    if (target_proc.returncode != 0):
        raise CaliperExecError('Command ' + str(target_cmd) + ' exited with ' + str(target_proc.returncode))
    if (query_proc.returncode != 0):
        raise CaliperExecError('Command ' + str(query_cmd) + ' exited with ' + str(query_proc.returncode))
    
    return query_out


def run_test(target_cmd, env):
    """ Execute a command, and return its output """

    target_proc = subprocess.Popen(target_cmd, env=env, stdout=subprocess.PIPE,stderr=subprocess.PIPE)
    proc_out, proc_err = target_proc.communicate()

    if (target_proc.returncode != 0):
        raise CaliperExecError('Command ' + str(target_cmd) + ' exited with ' + str(target_proc.returncode))

    return (proc_out,proc_err)


def get_snapshots_from_text(query_output):
    """ Translate expanded snapshots from `cali-query -e` output into list of dicts 

        Takes input in the form 

          attr1=x,attr2=y
          attr1=z
          ...

        and converts it to `[ { 'attr1' : 'x', 'attr2' : 'y' }, { 'attr1' : 'z' } ]`
    """

    snapshots = []

    for line in query_output.decode().splitlines():
        snapshots.append( { kv.partition('=')[0] : kv.partition('=')[2] for kv in line.split(',') } )

    return snapshots


def has_snapshot_with_attributes(snapshots, attrs):
    """ Test if a snapshot with the given subset of attributes exists """

    def dict_contains(a, b):
        for k in b.keys():
            if (a.get(k) != b.get(k)):
                return False
        return True

    for s in snapshots:
        if (dict_contains(s, attrs)):
            return True

    return False
    
def has_snapshot_with_keys(snapshots, keys):
    """ Test if a snapshot with the given subset of keys exists """
    return any( set(keys).issubset(set(s.keys())) for s in snapshots )

def get_snapshot_with_keys(snapshots, keys):
    """ Return the first snapshot that with the given subset of keys"""

    sk = set(keys)

    for s in snapshots:
        if (sk.issubset(set(s.keys()))):
            return s

    return None



#!@PYTHON_EXECUTABLE@
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import

__author__ = "Muhammad Haseeb"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Muhammad Haseeb"]
__license__ = "MIT"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"

try:
    import mpi4py  # noqa: F401
    from mpi4py import MPI  # noqa: F401
except ImportError:
    pass

import time
import unittest
import timemory as tim
from timemory import settings as settings
from timemory import component as comp


# --------------------------- tolerance variables -------------------------------------- #
# deltas for assert
deltatimer = 0.015
deltautil = 2.5


# --------------------------- helper functions ----------------------------------------- #
# check availability of a component
def check_available(component):
    return getattr(component, "available")


# compute fibonacci
def fibonacci(n):
    return n if n < 2 else (fibonacci(n - 1) + fibonacci(n - 2))


# cpu utilization for n milliseconds
def consume(n):
    # Python 3.7 has time_ns()
    try:
        # get current time in nsec
        now = time.time_ns()
        # try until time point
        while time.time_ns() < (now + (n * 1e6)):
            pass
    except AttributeError:
        now = 1000 * time.time()
        # try until time point
        while (1000 * time.time()) < (now + n):
            pass


# -------------------------- Timing Tests set ---------------------------------------- #
# Timing tests class
class TimemoryTimingTests(unittest.TestCase):
    # setup class: timemory settings
    @classmethod
    def setUpClass(self):
        settings.timing_units = "sec"
        settings.timing_precision = 9
        settings.json_output = True
        tim.init()  # need args? timemory is inited as soon as you import timemory
        settings.dart_output = True
        settings.dart_count = True
        settings.dart_count = 1
        settings.banner = False

    # tear down class: finalize
    @classmethod
    def tearDownClass(self):
        # tim.finalize()
        # tim.dmp.finalize()
        pass

    # ---------------------------------------------------------------------------------- #
    # test wall timer
    def test_wall_clock(self):
        """wall_timer"""
        if not comp.WallClock.available:
            raise unittest.SkipTest("[{}] not available".format("wall_clock"))

        obj = comp.WallClock(self.shortDescription())
        obj.push()
        obj.start()
        time.sleep(1)
        obj.stop()
        obj.pop()

        print("\n[{}]> result: {}".format(self.shortDescription(), obj.get()))

        # check for near equal
        self.assertAlmostEqual(1.0, obj.get(), delta=deltatimer)

    # ---------------------------------------------------------------------------------- #
    # test monotonic timer
    def test_monotonic_clock(self):
        """monotonic_timer"""
        if not comp.MonotonicClock.available:
            raise unittest.SkipTest(
                "[{}] not available".format("monotonic_clock")
            )

        obj = comp.MonotonicClock(self.shortDescription())
        obj.push()
        obj.start()
        time.sleep(1)
        obj.stop()
        obj.pop()

        print("\n[{}]> result: {}".format(self.shortDescription(), obj.get()))

        # check for near equal
        self.assertAlmostEqual(1.0, obj.get(), delta=deltatimer)

    # ---------------------------------------------------------------------------------- #
    # test monotonic timer raw
    def test_monotonic_raw_clock(self):
        """monotonic_raw_timer"""
        if not comp.MonotonicRawClock.available:
            raise unittest.SkipTest(
                "[{}] not available".format("monotonic_raw_clock")
            )

        obj = comp.MonotonicRawClock(self.shortDescription())
        obj.push()
        obj.start()
        time.sleep(1)
        obj.stop()
        obj.pop()

        print("\n[{}]> result: {}".format(self.shortDescription(), obj.get()))

        # check for near equal
        self.assertAlmostEqual(1.0, obj.get(), delta=deltatimer)

    # ---------------------------------------------------------------------------------- #
    # test system timer
    def test_system_clock(self):
        """system_timer"""
        if not comp.SysClock.available:
            raise unittest.SkipTest("[{}] not available".format("sys_clock"))

        obj = comp.SysClock(self.shortDescription())
        obj.push()
        obj.start()
        time.sleep(1)
        obj.stop()
        obj.pop()

        print("\n[{}]> result: {}".format(self.shortDescription(), obj.get()))

        # check for near equal
        self.assertAlmostEqual(0.0, obj.get(), delta=deltatimer)

    # ---------------------------------------------------------------------------------- #
    # test user timer
    def test_user_clock(self):
        """user_timer"""
        if not comp.UserClock.available:
            raise unittest.SkipTest("[{}] not available".format("user_clock"))

        obj = comp.UserClock(self.shortDescription())
        obj.push()
        obj.start()
        time.sleep(1)
        obj.stop()
        obj.pop()

        print("\n[{}]> result: {}".format(self.shortDescription(), obj.get()))

        # check for near equal
        self.assertAlmostEqual(0.0, obj.get(), delta=deltatimer)

    # ---------------------------------------------------------------------------------- #
    # test cpu timer
    def test_cpu_clock(self):
        """cpu_timer"""
        if not comp.CpuClock.available:
            raise unittest.SkipTest("[{}] not available".format("cpu_clock"))

        obj = comp.CpuClock(self.shortDescription())
        obj.push()
        obj.start()
        time.sleep(1)
        obj.stop()
        obj.pop()

        print("\n[{}]> result: {}".format(self.shortDescription(), obj.get()))

        # check for near equal
        self.assertAlmostEqual(0.0, obj.get(), delta=deltatimer)

    # ---------------------------------------------------------------------------------- #
    # test cpu utilization
    def test_cpu_util(self):
        """cpu_utilization"""
        if not comp.CpuUtil.available:
            raise unittest.SkipTest("[{}] not available".format("cpu_util"))

        obj = comp.CpuUtil(self.shortDescription())
        obj.push()
        obj.start()
        # work for 750ms
        consume(750)
        # sleep for 250ms
        time.sleep(0.25)
        obj.stop()
        obj.pop()

        print("\n[{}]> result: {}".format(self.shortDescription(), obj.get()))

        # check for near equal
        self.assertAlmostEqual(75.0, obj.get(), delta=deltautil)

    # ---------------------------------------------------------------------------------- #
    # test thread cpu clock
    def test_thread_cpu_clock(self):
        """thread_cpu_timer"""
        if not comp.ThreadCpuClock.available:
            raise unittest.SkipTest(
                "[{}] not available".format("thread_cpu_clock")
            )

        obj = comp.ThreadCpuClock(self.shortDescription())
        obj.push()
        obj.start()
        time.sleep(1)
        obj.stop()
        obj.pop()

        print("\n[{}]> result: {}".format(self.shortDescription(), obj.get()))

        # check for near equal
        self.assertAlmostEqual(0.0, obj.get(), delta=deltatimer)

    # ---------------------------------------------------------------------------------- #
    # test thread cpu utilization
    def test_thread_cpu_util(self):
        """thread_cpu_utilization"""
        if not comp.ThreadCpuUtil.available:
            raise unittest.SkipTest(
                "[{}] not available".format("thread_cpu_util")
            )

        obj = comp.ThreadCpuUtil(self.shortDescription())
        obj.push()
        obj.start()
        # work for 750ms
        consume(750)
        # sleep for 250ms
        time.sleep(0.25)
        obj.stop()
        obj.pop()

        print("\n[{}]> result: {}".format(self.shortDescription(), obj.get()))

        # check for near equal
        self.assertAlmostEqual(75.0, obj.get(), delta=deltautil)

    # ---------------------------------------------------------------------------------- #
    # test process_cpu_clock
    def test_process_cpu_clock(self):
        """process_cpu_timer"""
        if not comp.ProcessCpuClock.available:
            raise unittest.SkipTest(
                "[{}] not available".format("process_cpu_clock")
            )

        obj = comp.ProcessCpuClock(self.shortDescription())
        obj.push()
        obj.start()
        time.sleep(1)
        obj.stop()
        obj.pop()

        print("\n[{}]> result: {}".format(self.shortDescription(), obj.get()))

        # check for near equal
        self.assertAlmostEqual(0.0, obj.get(), delta=deltatimer)

    # ---------------------------------------------------------------------------------- #
    # test process cpu utilization
    def test_process_cpu_util(self):
        """process_cpu_utilization"""
        if not comp.ProcessCpuUtil.available:
            raise unittest.SkipTest(
                "[{}] not available".format("process_cpu_util")
            )

        obj = comp.ProcessCpuUtil(self.shortDescription())
        obj.push()
        obj.start()
        # work for 750ms
        consume(750)
        # sleep for 250ms
        time.sleep(0.25)
        obj.stop()
        obj.pop()

        print("\n[{}]> result: {}".format(self.shortDescription(), obj.get()))

        # check for near equal
        self.assertAlmostEqual(75.0, obj.get(), delta=deltautil)


# ----------------------------- main test runner -------------------------------------- #
# test runner
def run():
    # run all tests
    unittest.main()


if __name__ == "__main__":
    tim.initialize([__file__])
    run()



#!/usr/bin/env python
#
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" @file plotting/__init__.py
Plotting routines for timemory module
"""

from __future__ import absolute_import
from __future__ import division

import os
import sys
import json
import argparse
import traceback

__author__ = "Jonathan Madsen"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Jonathan Madsen"]
__license__ = "MIT"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"

from .plotting import (
    plot_maximums,
    plot,
    plot_all,
    plot_generic,
    read,
    plot_data,
    timemory_data,
    echo_dart_tag,
    add_plotted_files,
    make_output_directory,
    plot_parameters,
    plotted_files,
)

from .table import timem

__all__ = [
    "plotting",
    "plot",
    "plot_maximums",
    "plot_all",
    "plot_generic",
    "read",
    "plot_data",
    "timemory_data",
    "echo_dart_tag",
    "add_plotted_files",
    "make_output_directory",
    "plot_parameters",
    "plotted_files",
    "embedded_plot",
    "table",
    "timem",
]


def embedded_plot(
    _argv, _call_exit=False, _verbose=os.environ.get("TIMEMORY_VERBOSE", 0)
):
    """This is intended to be called from the embedded python interpreter"""
    _call_exit = False
    if _argv is None:
        _argv = sys.argv
        _call_exit = True

    try:
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-f", "--files", nargs="*", help="File input", type=str
        )
        parser.add_argument(
            "-d",
            "--display",
            required=False,
            action="store_true",
            help="Display plot",
            dest="display_plot",
        )
        parser.add_argument(
            "-t", "--titles", nargs="*", help="Plot titles", type=str
        )
        parser.add_argument(
            "-c",
            "--combine",
            required=False,
            action="store_true",
            help="Combined data into a single plot",
        )
        parser.add_argument(
            "-o",
            "--output-dir",
            help="Output directory",
            type=str,
            required=False,
        )
        parser.add_argument(
            "-e",
            "--echo-dart",
            help="echo Dart measurement for CDash",
            required=False,
            action="store_true",
        )
        parser.add_argument(
            "--min-percent",
            required=False,
            type=float,
            help="Exclude plotting below this percentage of maximum",
        )
        parser.add_argument(
            "--img-dpi",
            help="Image dots per sq inch",
            required=False,
            type=int,
        )
        parser.add_argument(
            "--img-size",
            help="Image dimensions",
            nargs=2,
            required=False,
            type=int,
        )
        parser.add_argument(
            "--img-type", help="Image type", required=False, type=str
        )
        parser.add_argument(
            "--plot-max",
            help="Plot the maximums from a set of inputs to <filename>",
            required=False,
            type=str,
            dest="plot_max",
        )
        parser.add_argument(
            "--log-x", help="Plot X-axis in a log scale", action="store_true"
        )
        parser.add_argument(
            "--font-size", help="Font size of y-axis labels", type=int
        )
        parser.add_argument(
            "--exit-on-failure",
            help="Call exit if failed",
            action="store_true",
        )
        parser.add_argument(
            "-T",
            "--table",
            help="Generate a table",
            action="store_true",
        )

        parser.set_defaults(display_plot=False)
        parser.set_defaults(combine=False)
        parser.set_defaults(output_dir=".")
        parser.set_defaults(echo_dart=False)
        parser.set_defaults(min_percent=plot_parameters.min_percent)
        parser.set_defaults(img_dpi=plot_parameters.img_dpi)
        parser.set_defaults(
            img_size=[
                plot_parameters.img_size["w"],
                plot_parameters.img_size["h"],
            ]
        )
        parser.set_defaults(img_type=plot_parameters.img_type)
        parser.set_defaults(plot_max="")
        parser.set_defaults(font_size=plot_parameters.font_size)

        args = parser.parse_args(_argv)

        do_plot_max = True if len(args.plot_max) > 0 else False

        params = plot_parameters(
            min_percent=args.min_percent,
            img_dpi=args.img_dpi,
            img_size={"w": args.img_size[0], "h": args.img_size[1]},
            img_type=args.img_type,
            log_xaxis=args.log_x,
            font_size=args.font_size,
        )

        if not args.table:
            if do_plot_max:
                if len(args.titles) != 1:
                    raise Exception("Error must provide one title")
            else:
                if len(args.titles) != 1 and len(args.titles) != len(
                    args.files
                ):
                    raise Exception(
                        "Error must provide one title or a title for each file"
                    )

        data = {}
        for i in range(len(args.files)):
            f = open(args.files[i], "r")
            _jdata = json.load(f)
            _cdata = _jdata["timemory"]

            if "timem" in _cdata:
                if args.table:
                    timem(_cdata, args.files[i], args)
                    continue
                _tdata = _cdata["timem"]
                nranks = len(_tdata)
                concurrency = 1  # always unknown
                _groups = {
                    "time": [
                        "wall_clock",
                        "user_clock",
                        "system_clock",
                        "cpu_clock",
                    ],
                    "utilization": [
                        "cpu_util",
                    ],
                    "memory": [
                        "peak_rss",
                        "page_rss",
                        "virtual_memory",
                    ],
                    "page_faults": [
                        "num_major_page_faults",
                        "num_minor_page_faults",
                    ],
                    "context_switches": [
                        "priority_context_switch",
                        "voluntary_context_switch",
                    ],
                    "io": [
                        "read_char",
                        "read_bytes",
                        "written_char",
                        "written_bytes",
                    ],
                    "hw": [
                        "papi_array",
                    ],
                }

                for cid, itr in _groups.items():
                    data = {}
                    unitr = ""
                    for k in itr:
                        if k in _tdata and "unit_repr" in _tdata[k]:
                            unitr = _tdata[k]["unit_repr"]
                            break

                    _dict = {
                        "cid": cid,
                        "mpi_size": nranks,
                        "concurrency": concurrency,
                        "units": unitr,
                        "ctype": cid,
                        "description": "",
                        "plot_params": params,
                    }

                    for j in range(0, nranks):
                        rdata = _tdata[j]

                        _idata = {}
                        for k in itr:
                            if k in rdata.keys():
                                _idata[k] = rdata[k]
                                if not unitr and "unit_repr" in rdata[k]:
                                    _dict["units"] = rdata[k]["unit_repr"]

                        if len(_idata) == 0:
                            continue
                        _data = read(_idata, **_dict)
                        _rtag = "" if nranks == 1 else "_{}".format(j)
                        _rtitle = (
                            "" if nranks == 1 else " (MPI rank: {})".format(j)
                        )

                        _data.filename = args.files[i].replace(".json", _rtag)
                        _data.filename = "{}_{}".format(_data.filename, cid)
                        if len(args.titles) == 1:
                            _data.title = args.titles[0] + _rtitle
                        else:
                            _data.title = args.titles[i] + _rtitle
                        if j not in data.keys():
                            data[j] = [_data]
                        else:
                            data[j] += [_data]

                    _pargs = {
                        "plot_params": params,
                        "display": args.display_plot,
                        "output_dir": args.output_dir,
                        "echo_dart": args.echo_dart,
                    }

                    if do_plot_max:
                        _pargs["combine"] = args.combine

                    for _rank, _data in data.items():
                        if do_plot_max:
                            plot_maximums(
                                args.plot_max, args.titles[0], _data, **_pargs
                            )
                        else:
                            plot(data=_data, **_pargs)

            else:
                for key, _json in _cdata.items():
                    nranks = (
                        int(_json["num_ranks"]) if "num_ranks" in _json else 1
                    )
                    concurrency = (
                        int(_json["concurrency"])
                        if "concurrency" in _json
                        else 1
                    )
                    ctype = _json["type"]  # collection type
                    cdesc = _json["description"]  # collection description
                    unitr = _json["unit_repr"]  # collection unit display repr
                    cid = " ".join(key.split("_")).title()
                    # roofl = _json["roofline"] if "roofline" in _json else None

                    _dict = {
                        "cid": cid,
                        "mpi_size": nranks,
                        "concurrency": concurrency,
                        "units": unitr,
                        "ctype": ctype,
                        "description": cdesc,
                        "plot_params": params,
                    }

                    rdata = _json["ranks"]
                    for j in range(0, len(rdata)):

                        _data = read(rdata[j], **_dict)
                        _rtag = "" if len(rdata) == 1 else "_{}".format(j)
                        _rtitle = (
                            ""
                            if len(rdata) == 1
                            else " (MPI rank: {})".format(j)
                        )

                        _data.filename = args.files[i].replace(".json", _rtag)
                        if len(args.titles) == 1:
                            _data.title = args.titles[0] + _rtitle
                        else:
                            _data.title = args.titles[i] + _rtitle
                        if j not in data.keys():
                            data[j] = [_data]
                        else:
                            data[j] += [_data]

            _pargs = {
                "plot_params": params,
                "display": args.display_plot,
                "output_dir": args.output_dir,
                "echo_dart": args.echo_dart,
            }

            if do_plot_max:
                _pargs["combine"] = args.combine

            for _rank, _data in data.items():
                if do_plot_max:
                    plot_maximums(
                        args.plot_max, args.titles[0], _data, **_pargs
                    )
                else:
                    plot(data=_data, **_pargs)

    except Exception as e:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback, limit=5)
        print("Exception - {}".format(e))
        if _call_exit or args.exit_on_failure:
            sys.exit(1)



# report / C config test

import unittest

import calipertest as cat

class CaliperReportTest(unittest.TestCase):
    """ Caliper TAU service test case """

    def test_tau(self):
        target_cmd = [ './ci_test_mpi_before_cali' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'tau',
            'CALI_LOG_VERBOSITY'     : '0',
            'TAU_PROFILE'            : '0',
            'TAU_TRACE'              : '0'
        }

        #query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        query_output = cat.run_test(target_cmd, caliper_config)
        #snapshots = cat.get_snapshots_from_text(query_output)

        #self.assertTrue(len(snapshots) == 5)

        #self.assertTrue(cat.has_snapshot_with_attributes(
            #snapshots, { 'iteration#fooloop': '3', 'count': '1' }))

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import sys
import argparse
# read in components and mangled enums
from timemory_types import components, mangled_enums, mangled_strings


def generate_if_statement(component, idx, indent_tabs=2, spaces=4, var="_hashmap"):
    """
    This function generates a case label for C++
    """
    enumeration = mangled_enums.get(component, component)
    component_options = mangled_strings.get(component, [component])
    if component not in component_options:
        component_options += [component]
    component_options.sort()

    tab = " " * spaces * indent_tabs
    statement = ""
    for comp in component_options:
        statement += '{}{}"{}", {}{},\n'.format(
            tab, '{', comp, enumeration.upper(), '}')
    return statement


if __name__ == "__main__":

    rbegin = "// GENERATE_SWITCH_REPLACE_BEGIN"
    rend = "// GENERATE_SWITCH_REPLACE_END"

    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--additional",
                        help="Additional components to produce case labels for", nargs='*')
    parser.add_argument("-i", "--input", help="input file for replacing",
                        type=str, default=None)
    parser.add_argument("-o", "--output", help="output components from file",
                        type=str, default=None)
    parser.add_argument("-e", "--exclude",
                        help="exclude components", action='store_true')
    parser.add_argument("-s", "--spaces-per-tab",
                        help="Number of spaces in a tab", type=int, default=4)
    parser.add_argument("-t", "--tabs-per-indent",
                        help="Number of tabs to indent", type=int, default=2)
    parser.add_argument("-r", "--replace", action='store_true',
                        help="Replace area between '{}' and '{}' in the output file".format(rbegin, rend))
    parser.add_argument("-c", "--config", help="alternative to -i/--input when replacing",
                        type=str, default=None)
    parser.add_argument("-v", "--var", help="Name of the map variable",
                        type=str, default="_instance")
    parser.add_argument("-I", "--iter-var", help="Name of the iteration variable",
                        type=str, default="itr")
    parser.add_argument("-V", "--verbose",
                        help="Enable verbosity", default=0, type=int)

    args = parser.parse_args()

    if args.replace and args.input is None and args.output is None:
        raise Exception(
            "-r/--replace requires specifying -o/--output or -i/--input")

    # substitution string for replacing
    subdata = None

    if args.input is not None:
        with open(args.input, 'r') as f:
            subdata = f.read()

    if args.config is not None:
        with open(args.config, 'r') as f:
            components = f.read().strip().split()

    if args.additional:
        components += args.additional

    if subdata is None and args.replace:
        if args.input is None:
            print("reading '{}' for replacement...".format(args.output))
            with open(args.output, 'r') as f:
                subdata = f.read()

    components.sort()
    if args.verbose > 0:
        print("timemory components: [{}]\n".format(", ".join(components)))

    component_options = []
    for component in components:
        component_options += mangled_strings.get(component, [component])
        if component not in component_options:
            component_options.append(component)
    component_options.sort()

    tab = " " * args.spaces_per_tab * args.tabs_per_indent
    btab = " " * args.spaces_per_tab * (args.tabs_per_indent - 1)
    outdata = "{}static component_hash_map_t _hashmap = {}\n".format(
        btab, '{')
    idx = 0
    for component in components:
        outdata += "{}".format(generate_if_statement(component, idx,
                                                     args.tabs_per_indent, args.spaces_per_tab,
                                                     args.var))
        idx += 1
    outdata = outdata.strip(",\n")
    outdata += '\n{}{};\n'.format(btab, '}')
    # outdata += "{}return _instance;\n{}{}\n".format(tab, btab, '};')

    message = '{}static auto errmsg = [](const std::string& {}) {} fprintf(stderr, "Unknown component label: %s{}", {}.c_str()); {};'.format(
        btab, args.iter_var, "{", ". Valid choices are: {}\\n".format(component_options), args.iter_var, "}")

    outdata += "\n{}".format(message)

    if subdata is not None:
        try:
            sbegin = subdata.find(rbegin) + len(rbegin) + 1
            send = subdata.find(rend)
            substr = subdata[sbegin:send]
            outdata += " " * args.spaces_per_tab * (args.tabs_per_indent + 1)
            if args.verbose > 1:
                print("sbegin = {}\nsend = {}\nsubstring:\n{}\n\noutdata:\n{}\n".format(
                      sbegin, send, substr, outdata))
            outdata = subdata.replace(substr, outdata)
            if args.verbose > 1:
                print("converted:\n{}\n".format(subdata))
        except Exception as e:
            print(e)
            raise

    if args.output is not None:
        with open(args.output, 'w') as f:
            f.write(outdata)
    else:
        print("{}".format(outdata))



# Tests for the textlog service

import unittest

import calipertest as cat

class CaliperTextlogTest(unittest.TestCase):
    """ Caliper textlog test cases """

    def test_textlog(self):
        target_cmd = [ './ci_test_macros' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'      : 'event,textlog',
            'CALI_TEXTLOG_TRIGGER'      : 'iteration#mainloop',
            'CALI_TEXTLOG_FORMATSTRING' : '%function% iteration: %[2]iteration#mainloop',
            'CALI_LOG_VERBOSITY'        : '0',
        }

        log_targets = [
            'main iteration:  0',
            'main iteration:  1',
            'main iteration:  2',
            'main iteration:  3'
        ]

        report_out,_ = cat.run_test(target_cmd, caliper_config)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

    def test_textlog_autostring(self):
        target_cmd = [ './ci_test_macros' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'      : 'event,textlog,timestamp',
            'CALI_TEXTLOG_TRIGGER'      : 'loop',
            'CALI_LOG_VERBOSITY'        : '0',
        }

        log_targets = [
            'loop=mainloop/fooloop'
        ]

        report_out,_ = cat.run_test(target_cmd, caliper_config)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

if __name__ == "__main__":
    unittest.main()



#!@PYTHON_EXECUTABLE@
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import

__author__ = "Muhammad Haseeb"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Muhammad Haseeb"]
__license__ = "MIT"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"

try:
    import mpi4py  # noqa: F401
    from mpi4py import MPI  # noqa: F401
except ImportError:
    pass

import os
import time
import unittest
import threading
import inspect
import numpy as np
import timemory as tim
from timemory import settings as settings

# --------------------------- test setup variables ----------------------------------- #


# --------------------------- helper functions ----------------------------------------- #

# check availability of a component
def check_available(component):
    return inspect.isclass(component)


# compute fibonacci
def fibonacci(n):
    return n if n < 2 else (fibonacci(n - 1) + fibonacci(n - 2))


# sleep for n nanosec
def do_sleep(n):
    time.sleep(n * 1e-9)


# cpu utilization for n nanosec
def consume(n):
    # Python 3.7 has time_ns()
    try:
        # get current time in nsec
        now = time.time_ns()
        # try until time point
        while time.time_ns() < (now + n):
            pass
    except AttributeError:
        now = 1e9 * time.time()
        # try until time point
        while (1e9 * time.time()) < (now + n):
            pass


# get auto_tuple config
def get_config(items=["wall_clock"]):
    return [getattr(tim.component, x) for x in items]


# -------------------------- Thottle Tests set ---------------------------------------- #
# Throttle tests class
class TimemoryThrottleTests(unittest.TestCase):
    # setup class: timemory settings
    @classmethod
    def setUpClass(self):
        # set up environment variables
        os.environ["TIMEMORY_VERBOSE"] = "1"
        os.environ["TIMEMORY_COLLAPSE_THREADS"] = "OFF"

        settings.parse()

        settings.verbose = 1
        settings.debug = False
        settings.json_output = True
        settings.mpi_thread = False
        settings.file_output = False
        settings.dart_output = True
        settings.dart_count = 1
        settings.banner = False

        tim.trace.init("wall_clock", False, "throttle_tests")

        self.nthreads = 1

    # Tear down class: finalize
    @classmethod
    def tearDownClass(self):
        # unset environment variables
        del os.environ["TIMEMORY_VERBOSE"]
        del os.environ["TIMEMORY_COLLAPSE_THREADS"]
        pass

    # ---------------------------------------------------------------------------------- #
    # test expect_true
    def test_expect_true(self):
        """expect_true"""
        settings.debug = False
        n = 2 * settings.throttle_count
        tim.trace.push("true")
        for i in range(n):
            tim.trace.push(self.shortDescription())
            tim.trace.pop(self.shortDescription())
        tim.trace.pop("true")

        self.assertTrue(tim.trace.is_throttled(self.shortDescription()))

    # ---------------------------------------------------------------------------------- #
    # test expect_false
    def test_expect_false(self):
        """expect_false"""
        settings.debug = False
        n = 2 * settings.throttle_count
        v = 2 * settings.throttle_value

        for i in range(n):
            tim.trace.push(self.shortDescription())
            consume(v)
            tim.trace.pop(self.shortDescription())

        self.assertFalse(tim.trace.is_throttled(self.shortDescription()))

    # ---------------------------------------------------------------------------------- #
    def test_region_serial(self):
        """region_serial"""
        settings.debug = False

        def _run(name):
            tim.region.push("rsthread")
            n = 8 * settings.throttle_count
            for i in range(n):
                tim.region.push(name)
                tim.region.pop(name)

            tim.region.pop("rsthread")

        for i in range(self.nthreads):
            _run(self.shortDescription())

        # print(tim.trace.is_throttled(self.shortDescription()))
        # print(tim.trace.is_throttled("thread"))

    # ---------------------------------------------------------------------------------- #
    # test region_multithreaded
    def test_region_multithreaded(self):
        """region_multithreaded"""
        settings.debug = False

        def _run(name):
            tim.region.push("rthread")
            n = 8 * settings.throttle_count
            for i in range(n):
                tim.region.push(name)
                tim.region.pop(name)

            tim.region.pop("rthread")

        threads = []

        for i in range(self.nthreads):
            thd = threading.Thread(target=_run, args=(self.shortDescription(),))
            thd.start()
            threads.append(thd)

        for itr in threads:
            itr.join()

    # ---------------------------------------------------------------------------------- #
    # test multithreaded
    def test_multithreaded(self):
        """multithreaded"""
        settings.debug = False

        # np.array of False
        is_throttled = np.full(self.nthreads, False)

        # _run function
        def _run(name, idx):
            name = "{}_{}".format(name, idx)
            n = 2 * settings.throttle_count
            v = 2 * settings.throttle_value
            if idx % 2 == 1:
                for i in range(n):
                    tim.trace.push(name)
                    consume(v)
                    tim.trace.pop(name)
            else:
                for i in range(n):
                    tim.trace.push(name)
                    tim.trace.pop(name)

            is_throttled[idx] = tim.trace.is_throttled(name)

        # thread handles
        threads = []

        # all threads
        tim.trace.push(self.shortDescription())

        # make new threads
        for i in range(self.nthreads):
            thd = threading.Thread(
                target=_run, args=(self.shortDescription(), i)
            )
            thd.start()
            threads.append(thd)

        # wait for join
        for itr in threads:
            itr.join()

        tim.trace.pop(self.shortDescription())

        # check assertion
        for i in range(self.nthreads):
            _answer = False if (i % 2 == 1) else True
            print("thread " + str(i) + " throttling: " + str(is_throttled[i]))
            self.assertTrue(is_throttled[i] == _answer)


# ----------------------------- main test runner -------------------------------------- #
# main runner
def run():
    # run all tests
    unittest.main()


if __name__ == "__main__":
    tim.initialize([__file__])
    run()



# Copyright (c) 2020, Lawrence Livermore National Security, LLC.
# See top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

class ReaderError(Exception):
    def __init__(self, message):
        self.message = message



from __future__ import print_function
from IPython.core.magic import Magics, magics_class, line_magic
from IPython.display import HTML, Javascript, display

"""
   File: roundtrip.py
   Purpose: Pass data between Jupyter Python cells and
   Javascript variables.
   Hatchet-specific.
"""


@magics_class
class Roundtrip(Magics):

    # Note to self: Custom magic classes MUST call parent's constructor
    def __init__(self, shell):
        super(Roundtrip, self).__init__(shell)
        self.id_number = 0
        # Clean up namespace function
        display(
            HTML(
                "<script>function cleanUp() { argList =[]; element = null; cell_idx = -1}</script>"
            )
        )

    inputType = {
        "js": "text/javascript",
        "csv": "text/csv",
        "html": "text/html",
        "json": "text/json",
        "css": "text/css",
    }
    codeMap = {}

    @line_magic
    def loadVisualization(self, line):
        # Get command line args for loading the vis
        args = line.split(" ")
        name = "roundtripTreeVis" + str(self.id_number)
        path = ""
        if '"' in args[0]:
            path = args[0].replace('"', "")
        elif "'" in args[0]:
            path = args[0].replace("'", "")
        else:
            # Path is a variable from the nb namespace
            path = self.shell.user_ns[args[0]]

        fileAndPath = ""
        if path[-1] == "/":
            fileAndPath = path + "roundtripTree.js"
        else:
            fileAndPath = path + "/roundtripTree.js"

        javascriptFile = open(fileAndPath).read()

        # Source input files
        # Set up the object to map input files to what javascript expects
        argList = "<script> var argList = []; var elementTop = null; var cell_idx = -1;</script>"

        displayObj = display(HTML(argList), display_id=True)

        args[1] = self.shell.user_ns[args[1]]
        displayObj.update(Javascript('argList.push("' + str(args[1]) + '")'))

        # Get curent cell id
        self.codeMap[name] = javascriptFile

        preRun = """
        // Grab current context
        elementTop = element.get(0);"""
        displayObj.update(Javascript(preRun))

        self.runVis(name, javascriptFile)
        self.id_number += 1

    def runVis(self, name, javascriptFile):
        name = "roundtripTreeVis" + str(self.id_number)
        header = (
            """
                  <div id=\""""
            + name
            + """\"></div>
                  <script>
                  elementTop.appendChild(document.getElementById('"""
            + str(name)
            + """'));
                  element = document.getElementById('"""
            + str(name)
            + """');"""
        )
        footer = """</script>"""
        display(HTML(header + javascriptFile + footer))

    @line_magic
    def fetchData(self, dest):
        # added eval() to 'execute' the JS list-as-string as a Python list

        hook = (
            """
                var holder = jsNodeSelected;
                holder = '"' + holder + '"';
                IPython.notebook.kernel.execute('"""
            + str(dest)
            + """ = '+ eval(holder));
                //console.log('"""
            + str(dest)
            + """ = '+ holder);
               """
        )

        display(Javascript(hook))

        return display(Javascript(hook))


# Function to make module loading possible
def load_ipython_extension(ipython):
    ipython.register_magics(Roundtrip)



# Copyright 2020-2021 The Regents of the University of California, through Lawrence
# Berkeley National Laboratory, and other Hatchet Project Developers. See the
# top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import json
import pandas as pd

from hatchet.graphframe import GraphFrame
from ..node import Node
from ..graph import Graph
from ..frame import Frame
from ..util.timer import Timer


class TimemoryReader:
    """Read in timemory JSON output"""

    def __init__(self, input, select=None, **_kwargs):
        """Arguments:
        input (str or file-stream or dict or None):
            Valid argument types are:

            1. Filename for a timemory JSON tree file
            2. Open file stream to one of these files
            3. Dictionary from timemory JSON tree

        select (list of str):
            A list of strings which match the component enumeration names, e.g. ["cpu_clock"].

        per_thread (boolean):
            Ensures that when applying filters to the graphframe, frames with
            identical name/file/line/etc. info but from different threads are not
            combined

        per_rank (boolean):
            Ensures that when applying filters to the graphframe, frames with
            identical name/file/line/etc. info but from different ranks are not
            combined
        """

        if isinstance(input, dict):
            self.graph_dict = input
        elif isinstance(input, str) and input.endswith("json"):
            with open(input) as f:
                self.graph_dict = json.load(f)
        elif not isinstance(input, str):
            self.graph_dict = json.loads(input.read())
        else:
            raise TypeError("input must be dict, json file, or string")
        self.default_metric = None
        self.name_to_hnode = {}
        self.name_to_dict = {}
        self.timer = Timer()
        self.metric_cols = []
        self.properties = {}
        self.include_tid = True
        self.include_nid = True
        # the per_thread and per_rank settings make sure that
        # squashing doesn't collapse the threads/ranks
        self.per_thread = _kwargs["per_thread"] if "per_thread" in _kwargs else False
        self.per_rank = _kwargs["per_rank"] if "per_rank" in _kwargs else False
        if select is None:
            self.select = select
        elif isinstance(select, list):

            if select:

                def _get_select(val):
                    if isinstance(val, str):
                        return val.lower()
                    elif callable(val):
                        return val().lower()
                    raise TypeError(
                        "Items in select must be string or callable: {}".format(
                            type(val).__name__
                        )
                    )

                self.select = [_get_select(v) for v in select]
        else:
            raise TypeError("select must be None or list of string")

    def create_graph(self):
        """Create graph and dataframe"""

        list_roots = []
        node_dicts = []
        graph_dict = self.graph_dict

        def remove_keys(_dict, _keys):
            """Remove keys from dictionary"""
            if isinstance(_keys, str):
                if _keys in _dict:
                    del _dict[_keys]
            else:
                for _key in _keys:
                    _dict = remove_keys(_dict, _key)
            return _dict

        def patch_keys(_dict, _extra):
            """Add a suffix to dictionary keys"""
            _tmp = {}
            for key, itr in _dict.items():
                _tmp["{}{}".format(key, _extra)] = itr
            return _tmp

        def add_metrics(_dict):
            """Add any keys to metric_cols which don't already exist"""
            for key, itr in _dict.items():
                if key not in self.metric_cols:
                    self.metric_cols.append(key)

        def process_regex(_data):
            """Process the regex data for func/file/line info"""
            _tmp = {}
            if _data is not None and len(_data.groups()) > 0:
                for _key in ("head", "func", "file", "line", "tail"):
                    try:
                        _val = _data.group(_key)
                        if _val:
                            _tmp[_key] = _val
                    except Exception:
                        pass
            return _tmp if _tmp else None

        def perform_regex(_itr):
            """Performs a search for standard configurations of function + file + line"""
            import re

            _tmp = None
            for _pattern in [
                # [func][file]
                r"(^\[)(?P<func>.*)(\]\[)(?P<file>.*)(\]$)",
                # label  [func/file:line]
                r"(?P<head>.+)([ \t]+)\[(?P<func>\S+)([/])(?P<file>\S+):(?P<line>[0-9]+)\]$",
                # func@file:line/tail
                # func/file:line/tail
                r"(?P<func>\S+)([@/])(?P<file>\S+):(?P<line>[0-9]+)[/]*(?P<tail>.*)",
                # func@file/tail
                # func/file/tail
                r"(?P<func>\S+)([@/])(?P<file>\S+)([/])(?P<tail>.*)",
                # func:line/tail
                r"(?P<func>\S+):(?P<line>[0-9]+)([/]*)(?P<tail>.*)",
            ]:
                _tmp = process_regex(re.search(_pattern, _itr))
                if _tmp:
                    break
            return _tmp if _tmp else None

        def get_keys(_prefix):
            """Get the standard set of dictionary entries.
            Also, parses the prefix for func-file-line info
            which is typically in the form:
                <FUNC>@<FILE>:<LINE>/...
                <FUNC>/<FILE>:<LINE>/...
                <SOURCE>    [<FUNC>/<FILE>:<LINE>]
            """
            _keys = {
                "type": "region",
                "name": _prefix,
            }
            _extra = {"file": "<unknown>", "line": "0"}
            _pdict = perform_regex(_prefix)
            if _pdict is not None:
                if "head" in _pdict:
                    _keys["name"] = _pdict["head"].rstrip()
                    _extra["line"] = _pdict["line"]
                    _extra["file"] = _pdict["file"]
                else:
                    _keys["name"] = _pdict["func"]
                    _extra["file"] = (
                        _pdict["file"] if "file" in _pdict else "<unknown file>"
                    )
                    if "line" in _pdict:
                        _extra["line"] = _pdict["line"]
                    if "tail" in _pdict:
                        _keys["name"] = "{}/{}".format(_keys["name"], _pdict["tail"])
            return (_keys, _extra)

        def get_md_suffix(_obj):
            """Gets a multi-dimensional suffix"""
            _ret = []
            if isinstance(_obj, str):
                _ret = [_obj]
            elif isinstance(_obj, dict):
                for _key, _item in _obj.items():
                    _ret.append(_key.strip().replace(" ", "-").replace("_", "-"))
            elif isinstance(_obj, list) or isinstance(_obj, tuple):
                for _item in _obj:
                    _ret.append(_item.strip().replace(" ", "-").replace("_", "-"))
            return _ret

        def get_md_entries(_obj, _suffix):
            """Gets a multi-dimensional entries"""
            _ret = {}
            for _key, _item in _obj.items():
                for i, (k, v) in enumerate(_item.items()):
                    _ret["{}.{}".format(_key, _suffix[i])] = v
            return _ret

        def collapse_ids(_obj):
            """node/rank/thread id may be int, array of ints, or None.
            This function generates a consistent form which is not numerical to
            avoid groupby(...).sum() from producing something nonsensical but
            the entry is still hashable
            """
            if isinstance(_obj, list):
                return ",".join([f"{x}" for x in _obj]).strip(",")
            elif _obj is not None:
                return f"{_obj}"
            return None

        def parse_node(_key, _dict, _hparent, _rank):
            """Create node_dict for one node and then call the function
            recursively on all children.
            """

            # If the hash is zero, that indicates that the node
            # is a dummy for the root or is used for synchronizing data
            # between multiple threads
            if _dict["node"]["hash"] == 0:
                if "children" in _dict:
                    for _child in _dict["children"]:
                        parse_node(_key, _child, _hparent, _rank)
                return

            _prop = self.properties[_key]
            _keys, _extra = get_keys(_dict["node"]["prefix"])

            # by placing the thread-id or rank-id in _keys, the hash
            # for the Frame(_keys) effectively circumvent Hatchet's
            # default behavior of combining similar thread/rank entries
            _tid_dict = _keys if self.per_thread else _extra
            _rank_dict = _keys if self.per_rank else _extra

            # handle the rank
            _rank_dict["rank"] = collapse_ids(_rank)
            if _rank_dict["rank"] is None:
                del _rank_dict["rank"]
                self.include_nid = False

            # extract some relevant data
            _tid_dict["tid"] = collapse_ids(_dict["node"]["tid"])
            _extra["pid"] = collapse_ids(_dict["node"]["pid"])
            _extra["count"] = _dict["node"]["inclusive"]["entry"]["laps"]

            # this is the name for the metrics
            _labels = None if "type" not in _prop else _prop["type"]
            # if the labels are not a single string, they are multi-dimensional
            _md = True if not isinstance(_labels, str) else False

            # create the node
            _hnode = Node(Frame(_keys), _hparent)

            # remove some spurious data from inclusive/exclusive stats
            _remove = ["cereal_class_version", "count"]
            _inc_stats = remove_keys(_dict["node"]["inclusive"]["stats"], _remove)
            _exc_stats = remove_keys(_dict["node"]["exclusive"]["stats"], _remove)

            # if multi-dimensions, create alternative "sum.<...>", etc. labels + data
            if _md:
                _suffix = get_md_suffix(_labels)
                _exc_stats = get_md_entries(_exc_stats, _suffix)
                _inc_stats = get_md_entries(_inc_stats, _suffix)

            # add ".inc" to the end of every column that represents an inclusive stat
            _inc_stats = patch_keys(_inc_stats, ".inc")
            _exc_stats = patch_keys(_exc_stats, "")

            # add the inclusive and exclusive columns to the list of relevant column names
            add_metrics(_exc_stats)
            add_metrics(_inc_stats)

            # add an entry for the node
            node_dicts.append(
                dict({"node": _hnode, **_keys}, **_extra, **_exc_stats, **_inc_stats)
            )

            # set the node as the root or as a child
            if _hparent is None:
                list_roots.append(_hnode)
            else:
                _hparent.add_child(_hnode)

            # recursion
            if "children" in _dict:
                for _child in _dict["children"]:
                    parse_node(_key, _child, _hnode, _rank)

        def eval_graph(_key, _dict, _rank):
            """Evaluate the entry and determine if it has relevant data.
            If the hash is zero, that indicates that the node
            is a dummy for the root or is used for synchronizing data
            between multiple threads
            """
            _nchild = len(_dict["children"])
            if _nchild == 0:
                print("Skipping {}...".format(_key))
                return
            # if _rank is not None:
            #    print("Adding {} for rank {}...".format(_key, _rank))
            # else:
            #    print("Adding {}...".format(_key))

            if _dict["node"]["hash"] > 0:
                parse_node(_key, _dict, None, _rank)
            elif "children" in _dict:
                # call for all children
                for child in _dict["children"]:
                    parse_node(_key, child, None, _rank)

        def read_graph(_key, _itr, _offset):
            """The layout of the graph at this stage
            is subject to slightly different structures
            based on whether distributed memory parallelism (DMP)
            (e.g. MPI, UPC++) was supported and active
            """
            for i in range(len(_itr)):
                _dict = _itr[i]
                _idx = None if _offset is None else i + _offset
                if isinstance(_dict, list):
                    for j in range(len(_dict)):
                        eval_graph(_key, _dict[j], _idx)
                else:
                    eval_graph(_key, _dict, _idx)

        def read_properties(_dict, _key, _itr):
            """Read in the properties for a component. This
            contains information on the type of the component,
            a description, a unit_value relative to the
            standard, a unit label, whether the data is
            only relevant per-thread, the number of MPI and/or
            UPC++ ranks (some results can theoretically use
            both UPC++ and MPI), the number of threads in
            the application, and the total number of processes
            """
            if _key not in _dict:
                _dict[_key] = {}
            try:
                _dict[_key]["properties"] = remove_keys(
                    itr["properties"], "cereal_class_version"
                )
            except KeyError:
                pass
            for k in (
                "type",
                "description",
                "unit_value",
                "unit_repr",
                "thread_scope_only",
                "mpi_size",
                "upcxx_size",
                "thread_count",
                "process_count",
            ):
                if k not in _dict[_key] or _dict[_key][k] is None:
                    if k in itr:
                        _dict[_key][k] = itr[k]
                    else:
                        _dict[_key][k] = None

        for key, itr in graph_dict["timemory"].items():
            # strip out the namespace if provided
            key = key.replace("tim::", "").replace("component::", "").lower()
            # check for selection
            if self.select is not None and key not in self.select:
                continue
            # read in properties
            read_properties(self.properties, key, itr)
            # if no DMP supported
            if "graph" in itr:
                print("Reading graph...")
                read_graph(key, itr["graph"], None)
            else:
                # read in MPI results
                if "mpi" in itr:
                    # print("Reading MPI...")
                    read_graph(key, itr["mpi"], 0)
                # if MPI and UPC++, report ranks
                # offset by MPI_Comm_size
                _offset = self.properties[key]["mpi_size"]
                _offset = 0 if _offset is None else int(_offset)
                if "upc" in itr:
                    print("Reading UPC...")
                    read_graph(key, itr["upc"], _offset)
                elif "upcxx" in itr:
                    print("Reading UPC++...")
                    read_graph(key, itr["upcxx"], _offset)

        # find any columns where the entries are None or "null"
        non_null = {}
        for itr in node_dicts:
            for key, item in itr.items():
                if key not in non_null:
                    non_null[key] = False
                if item is not None:
                    if not isinstance(item, str):
                        non_null[key] = True
                    elif isinstance(item, str) and item != "null":
                        non_null[key] = True

        # find any columns where the entries are all "null"
        for itr in node_dicts:
            for key, item in non_null.items():
                if not item:
                    del itr[key]

        # create the graph of the roots
        graph = Graph(list_roots)
        graph.enumerate_traverse()

        # separate out the inclusive vs. exclusive columns
        exc_metrics = []
        inc_metrics = []
        for column in self.metric_cols:
            if column.endswith(".inc"):
                inc_metrics.append(column)
            else:
                exc_metrics.append(column)

        # set the default metric
        if self.default_metric is None:
            if len(exc_metrics) > 0:
                self.default_metric = exc_metrics[0]
            elif len(inc_metrics) > 0:
                self.default_metric = inc_metrics[0]
            else:
                self.default_metric = "sum"

        indices = ["node"]
        # this attempt at MultiIndex does not work
        # if self.include_nid:
        #    indices.append("nid")
        # if self.include_tid:
        #    indices.append("tid")
        dataframe = pd.DataFrame(data=node_dicts)
        dataframe.set_index(indices, inplace=True)
        dataframe.sort_index(inplace=True)

        return GraphFrame(
            graph, dataframe, exc_metrics, inc_metrics, self.default_metric
        )

    def read(self):
        """Read timemory json."""
        return self.create_graph()



#!@PYTHON_EXECUTABLE@

"""Example
@PYTHON_EXECUTABLE@ -m timemory.profiler -b -m 10 -- ./@FILENAME@
@PYTHON_EXECUTABLE@ -m timemory.line_profiler -b -v -- ./@FILENAME@
@PYTHON_EXECUTABLE@ -m timemory.trace -b -- ./@FILENAME@
"""

import sys
import numpy as np


@profile  # noqa: F821
def fib(n):
    return n if n < 2 else (fib(n - 1) + fib(n - 2))


@profile  # noqa: F821
def inefficient(n):
    a = 0
    for i in range(n):
        a += i
        for j in range(n):
            a += j
    arr = np.arange(a * n * n * n, dtype=np.double)
    return arr.sum()


def run(nfib):
    ret = 0
    ret += fib(nfib)
    ret += fib(nfib % 5 + 1)
    ret += inefficient(nfib)
    return ret


if __name__ == "__main__":
    nfib = int(sys.argv[1]) if len(sys.argv) > 1 else 20
    ans = run(nfib)
    print("Answer = {}".format(ans))



#!/usr/bin/env python

import sys
import argparse
# read in components and mangled enums
from timemory_types import components, mangled_enums


def generate_enum(component, indent):
    """
    This function generates a type trait label for C++
    """
    enumeration = mangled_enums.get(component, component)
    return "{}.value({}{}{}, {})".format(" " * indent, '"', enumeration, '"',
                                       enumeration.upper())


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-V", "--verbose", help="Enable verbosity", default=0, type=int)
    parser.add_argument("-i", "--indent", help="Indentation", default=8, type=int)
    args = parser.parse_args()

    if args.verbose > 0:
        print("timemory components: [{}]\n".format(", ".join(components)))

    components.sort()
    outdata = "components_enum\n"
    for component in components:
        outdata += "{}\n".format(generate_enum(component, args.indent))

    outdata = outdata.strip("\n")
    outdata += ";"
    print("{}".format(outdata))



#!/usr/bin/env python

import hatchet as ht
from sys import version_info as python_version


if __name__ == "__main__":
    # Path to pstats file.
    # A pstats file produced with a particualar version of python (2 or 3) must be read in with that version
    if python_version[0] == 2:
        pstats_file = "../../../hatchet/tests/data/cprofile-hatchet-pstats/cprofile-cycle-py2.pstats"
    elif python_version[0] == 3:
        pstats_file = (
            "../../../hatchet/tests/data/cprofile-hatchet-pstats/cprofile-cycle.pstats"
        )

    # Use hatchet's ``from_cprofile`` API to read in the pstats file. The result
    # is stored into Hatchet's GraphFrame.
    gf = ht.GraphFrame.from_cprofile(pstats_file)

    # Printout the DataFrame component of the GraphFrame.
    print(gf.dataframe)

    # Printout the graph component of the GraphFrame.
    # Because no metric parameter is specified, ``time`` is used by default.
    print(gf.tree())



#!/usr/bin/env python

import sys
import argparse
# read in components and mangled enums
from timemory_types import components, mangled_enums, mangled_strings


def generate_get_enum(component):
    """
    This function generates a case label for C++
    """
    enumeration = mangled_enums.get(component, component).upper()
    strings = ["{}".format(component)] + mangled_strings.get(component, [])
    if len(strings) == 1:
        strings += ['']

    quoted = ('"{}"'.format(x) for x in strings)

    return "TIMEMORY_PROPERTY_SPECIALIZATION({}, {}, {})".format(
           component, enumeration, ", ".join(quoted))


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--additional", nargs='*',
                        help="Additional components to produce case labels for")
    parser.add_argument("-e", "--exclude",
                        help="exclude components", action='store_true')
    parser.add_argument("-s", "--spaces-per-tab",
                        help="Number of spaces in a tab", type=int, default=4)
    parser.add_argument("-t", "--tabs-per-indent",
                        help="Number of tabs to indent", type=int, default=1)
    parser.add_argument("-V", "--verbose",
                        help="Enable verbosity", default=0, type=int)

    args = parser.parse_args()

    if args.additional:
        components += args.additional

    components.sort()
    if args.verbose > 0:
        print("timemory components: [{}]\n".format(", ".join(components)))
    outdata = ""
    for component in components:
        outdata += "{}\n".format(generate_get_enum(component))
        outdata += "\n//{}//\n\n".format("-" * 86)
    outdata = outdata.strip("\n")
    print("{}".format(outdata))



#!@PYTHON_EXECUTABLE@
#
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import

""" @file __main__.py
Command line execution for tracer
"""

import os
import sys
import argparse
import traceback

PY3 = sys.version_info[0] == 3


# Python 3.x compatibility utils: execfile
try:
    execfile
except NameError:
    # Python 3.x doesn't have 'execfile' builtin
    import builtins

    exec_ = getattr(builtins, "exec")

    def execfile(filename, globals=None, locals=None):
        with open(filename, "rb") as f:
            exec_(compile(f.read(), filename, "exec"), globals, locals)


def find_script(script_name):
    """Find the script.

    If the input is not a file, then $PATH will be searched.
    """
    if os.path.isfile(script_name):
        return script_name
    path = os.getenv("PATH", os.defpath).split(os.pathsep)
    for dir in path:
        if dir == "":
            continue
        fn = os.path.join(dir, script_name)
        if os.path.isfile(fn):
            return fn

    sys.stderr.write("Could not find script %s\n" % script_name)
    raise SystemExit(1)


def parse_args(args=None):
    """Parse the arguments"""

    if args is None:
        args = sys.argv

    from ..libs.libpytimemory.trace import config as _tracer_config

    def str2bool(v):
        if isinstance(v, bool):
            return v
        if v.lower() in ("yes", "true", "t", "y", "1"):
            return True
        elif v.lower() in ("no", "false", "f", "n", "0"):
            return False
        else:
            raise argparse.ArgumentTypeError("Boolean value expected.")

    parser = argparse.ArgumentParser(add_help=True)
    parser.add_argument(
        "-c",
        "--components",
        nargs="+",
        default=["wall_clock"],
        help="List of components",
    )
    parser.add_argument(
        "-d",
        "--output-dir",
        default=None,
        type=str,
        help="Output path of results",
    )
    parser.add_argument(
        "-b",
        "--builtin",
        action="store_true",
        help="Put 'trace' in the builtins. Use 'trace.enable()' and "
        "'trace.disable()' in your code to turn it on and off, or "
        "'@trace' to decorate a single function, or 'with trace:' "
        "to trace a single section of code.",
    )
    parser.add_argument(
        "-s",
        "--setup",
        default=None,
        help="Code to execute before the code to trace",
    )
    parser.add_argument(
        "-a",
        "--include-args",
        type=str2bool,
        nargs="?",
        const=True,
        default=_tracer_config.include_args,
        help="Encode the argument values",
    )
    parser.add_argument(
        "-l",
        "--include-line",
        type=str2bool,
        nargs="?",
        const=True,
        default=_tracer_config.include_line,
        help="Encode the function line number",
    )
    parser.add_argument(
        "-f",
        "--include-file",
        type=str2bool,
        nargs="?",
        const=True,
        default=_tracer_config.include_filename,
        help="Encode the function filename",
    )
    parser.add_argument(
        "-F",
        "--full-filepath",
        type=str2bool,
        nargs="?",
        const=True,
        default=_tracer_config.full_filepath,
        help="Encode the full function filename (instead of basename)",
    )
    parser.add_argument(
        "--skip-funcs",
        type=str,
        nargs="+",
        default=_tracer_config.skip_functions,
        help="Filter out any entries with these function names",
    )
    parser.add_argument(
        "--skip-files",
        type=str,
        nargs="+",
        default=_tracer_config.skip_filenames,
        help="Filter out any entries from these files",
    )
    parser.add_argument(
        "--only-funcs",
        type=str,
        nargs="+",
        default=_tracer_config.only_functions,
        help="Select only entries with these function names",
    )
    parser.add_argument(
        "--only-files",
        type=str,
        nargs="+",
        default=_tracer_config.only_filenames,
        help="Select only entries from these files",
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        type=int,
        default=_tracer_config.verbosity,
        help="Logging verbosity",
    )

    return parser.parse_known_args(args)


def get_value(env_var, default_value, dtype, arg=None):
    if arg is not None:
        return dtype(arg)
    else:
        val = os.environ.get(env_var)
        if val is None:
            os.environ[env_var] = "{}".format(default_value)
            return dtype(default_value)
        else:
            return dtype(val)


def run(tracer, cmd):
    if len(cmd) == 0:
        return

    progname = cmd[0]
    sys.path.insert(0, os.path.dirname(progname))
    with open(progname, "rb") as fp:
        code = compile(fp.read(), progname, "exec")

    import __main__

    dict = __main__.__dict__
    print("code: {} {}".format(type(code).__name__, code))
    globs = {
        "__file__": progname,
        "__name__": "__main__",
        "__package__": None,
        "__cached__": None,
        **dict,
    }

    tracer.runctx(code, globs, None)


def main():
    """Main function"""

    opts = None
    argv = None
    if "--" in sys.argv:
        _idx = sys.argv.index("--")
        _argv = sys.argv[(_idx + 1) :]  # noqa: E203
        opts, argv = parse_args(sys.argv[:_idx])
        argv = _argv
    else:
        if "-h" in sys.argv or "--help" in sys.argv:
            opts, argv = parse_args()
        else:
            argv = sys.argv[1:]
            opts, discard = parse_args()
            if len(argv) == 0 or not os.path.isfile(argv[0]):
                raise RuntimeError(
                    "Could not determine input script. Use '--' before "
                    "the script and its arguments to ensure correct parsing. \nE.g. "
                    "python -m timemory.trace -- ./script.py"
                )

    from ..libs.libpytimemory import initialize
    from ..libs.libpytimemory import settings

    if os.path.isfile(argv[0]):
        argv[0] = os.path.realpath(argv[0])

    settings.output_path = "timemory-{}-trace-output".format(
        os.path.basename(argv[0]).strip().replace("_", "-").replace("--", "-")
    )
    initialize(argv)

    from ..libs.libpytimemory.trace import config as _tracer_config

    _tracer_config.include_args = opts.include_args
    _tracer_config.include_line = opts.include_line
    _tracer_config.include_filename = opts.include_file
    _tracer_config.full_filepath = opts.full_filepath
    _tracer_config.skip_functions = opts.skip_funcs
    _tracer_config.skip_filenames = opts.skip_files
    _tracer_config.only_functions = opts.only_funcs
    _tracer_config.only_filenames = opts.only_files
    _tracer_config.verbosity = opts.verbosity

    print("[timemory]> tracing: {}".format(argv))

    sys.argv[:] = argv
    if opts.setup is not None:
        # Run some setup code outside of the tracer. This is good for large
        # imports.
        setup_file = find_script(opts.setup)
        __file__ = setup_file
        __name__ = "__main__"
        # Make sure the script's directory is on sys.path
        sys.path.insert(0, os.path.dirname(setup_file))
        ns = locals()
        execfile(setup_file, ns, ns)

    from . import Tracer, FakeTracer

    output_path = get_value(
        "TIMEMORY_OUTPUT_PATH", settings.output_path, str, opts.output_dir
    )
    settings.output_path = output_path

    # if len(argv) > 1 and argv[0] != "--"):
    #    raise RuntimeError("Specify command to run with -- <command-to-run>")
    script_file = find_script(sys.argv[0])
    __file__ = script_file
    __name__ = "__main__"
    # Make sure the script's directory is on sys.path
    sys.path.insert(0, os.path.dirname(script_file))

    tracer = Tracer(opts.components)
    fake = FakeTracer()

    if PY3:
        import builtins
    else:
        import __builtin__ as builtins

    builtins.__dict__["trace"] = tracer
    builtins.__dict__["notrace"] = fake
    builtins.__dict__["profile"] = tracer
    builtins.__dict__["noprofile"] = fake

    try:
        try:
            if not opts.builtin:
                tracer.start()
            tracer._unset += 1
            execfile_ = execfile
            ns = locals()
            if opts.builtin:
                execfile(script_file, ns, ns)
            else:
                tracer.runctx(
                    "execfile_(%r, globals())" % (script_file,), ns, ns
                )
        except (KeyboardInterrupt, SystemExit):
            pass
        finally:
            if not opts.builtin:
                tracer.stop()
            tracer._unset -= 1
            del tracer
            del fake
    except Exception as e:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback, limit=10)
        print("Exception - {}".format(e))


if __name__ == "__main__":
    main()
    from ..libs.libpytimemory import finalize

    finalize()



# Sampler/symbollookup tests

import json
import unittest

import calipertest as cat

class CaliperSamplerTest(unittest.TestCase):
    """ Caliper sampler test case """

    def test_sampler_symbollookup(self):
        target_cmd = [ './ci_test_macros', '5000' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_SERVICES_ENABLE'   : 'sampler:symbollookup:trace:recorder',
            'CALI_SYMBOLLOOKUP_LOOKUP_FILE'   : 'true',
            'CALI_SYMBOLLOOKUP_LOOKUP_LINE'   : 'true',
            'CALI_SYMBOLLOOKUP_LOOKUP_MODULE' : 'true',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) > 1)

        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, { 'cali.sampler.pc', 
                         'source.function#cali.sampler.pc', 
                         'source.file#cali.sampler.pc',
                         'source.line#cali.sampler.pc',
                         'sourceloc#cali.sampler.pc',
                         'module#cali.sampler.pc',
                         'function', 'loop' }))

    def test_hatchet_sample_profile_lookup(self):
        target_cmd = [ './ci_test_macros', '5000', 'hatchet-sample-profile(use.mpi=false,output=stdout,sample.callpath=false,lookup.sourceloc=true,lookup.module=true)' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY' : '0'
        }

        obj = json.loads( cat.run_test(target_cmd, caliper_config)[0] )

        self.assertEqual(set(obj['columns']), { 'count', 'time', 'sourceloc#cali.sampler.pc', 'module#cali.sampler.pc', 'path' } )

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import hatchet as ht


if __name__ == "__main__":
    # Path to HPCToolkit database directory.
    dirname = "../../../hatchet/tests/data/hpctoolkit-cpi-database"

    # Use hatchet's ``from_hpctoolkit`` API to read in the HPCToolkit database.
    # The result is stored into Hatchet's GraphFrame.
    gf = ht.GraphFrame.from_hpctoolkit(dirname)

    # Printout the DataFrame component of the GraphFrame.
    print(gf.dataframe)

    # Printout the graph component of the GraphFrame.
    # Use "time (inc)" as the metric column to be displayed
    print(gf.tree(metric_column="time (inc)"))



#!@PYTHON_EXECUTABLE@
#
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

""" @file __main__.py
Command line execution for roofline plotting library
"""

import os
import sys
import json
import argparse
import traceback


# error code
_errc = 0


def parse_args(add_run_args=False):
    parser = argparse.ArgumentParser(
        add_help=False,
        prog="timemory.roofline",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""\
Examples:
  - Perform a Roofline analysis with default parameters
        python -m timemory.roofline -- ./application
  - Perform a Roofline analysis with user-defined dimensions: 1000x800, DPI = 50
        python -m timemory.roofline -P 1000 800 50 -- ./application
                                     """,
    )

    parser.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="{} [OPTIONS [OPTIONS...]] -- <OPTIONAL COMMAND TO EXECUTE>".format(
            sys.argv[0]
        ),
    )
    parser.add_argument(
        "-d", "--display", action="store_true", help="Display plot"
    )
    parser.add_argument(
        "-o", "--output-file", type=str, help="Output file", default="roofline"
    )
    parser.add_argument(
        "-D",
        "--output-dir",
        type=str,
        help="Output directory",
        default=os.getcwd(),
    )
    lbandwidth = ["l1", "l2", "l3", "dram"]
    parser.add_argument(
        "-b",
        "--bandwidth",
        type=str,
        help="Bandwidth type - available optionss are: "
        + ", ".join(lbandwidth)
        + ', "dram" as default.',
        choices=lbandwidth,
        metavar="",
        dest="bandwidths",
        default=["dram"],
    )
    parser.add_argument(
        "-tb",
        "--txn_bandwidth",
        type=float,
        metavar="",
        help="GPU Instruction Roofline transaction bandwidth peak: L1, L2, DRAM. "
        + 'Three arguments are expected. "437.5 93.6 25.9" (NVIDIA V100 values)'
        + " as default.",
        dest="txns_bandwidths",
        default=[437.5, 93.6, 25.9],
        nargs=3,
    )
    parser.add_argument(
        "-iP",
        "--inst_peak",
        type=float,
        help="GPU Instruction peak (per warp) in GIPS (NVIDIA V100 peak set by default)",
        dest="inst_peak",
        default=[489.60],
        nargs=1,
    )
    parser.add_argument(
        "--format", type=str, help="Image format", default="png"
    )
    parser.add_argument(
        "-T",
        "--title",
        type=str,
        dest="title",
        help='Title for the plot, "Roofline" as default.',
        default="Roofline",
    )
    parser.add_argument(
        "-P",
        "--plot-dimensions",
        type=int,
        metavar="",
        help="Image dimensions: Width, Height, DPI. Three arguments are expected."
        + ' "1600 1200 100" as default.',
        default=[1600, 1200, 100],
        nargs=3,
    )
    parser.add_argument("-R", "--rank", type=int, help="MPI Rank", default=None)
    parser.add_argument(
        "-v", "--verbose", type=int, help="Verbosity", default=None
    )
    parser.add_argument(
        "-e",
        "--echo-dart",
        action="store_true",
        help="Echo image as DartMeasurementFile",
    )
    ltype = [
        "cpu_roofline",
        "cpu_roofline_sp",
        "cpu_roofline_dp",
        "gpu_roofline",
        "gpu_roofline_hp",
        "gpu_roofline_sp",
        "gpu_roofline_dp",
        "gpu_roofline_inst",
    ]
    parser.add_argument(
        "-t",
        "--rtype",
        type=str,
        choices=ltype,
        help="Roofline type - available optionss are: "
        + ", ".join(ltype)
        + ', "cpu_roofline" as default.',
        metavar="",
        default="cpu_roofline",
    )

    if add_run_args:
        parser.add_argument(
            "-p",
            "--preload",
            help="Enable preloading libtimemory.so",
            action="store_true",
        )
        parser.add_argument(
            "-k",
            "--keep-going",
            help="Continue despite execution errors",
            action="store_true",
        )
        parser.add_argument(
            "-r",
            "--rerun",
            help="Re-run this mode and not the other",
            type=str,
            choices=["ai", "op"],
            default=None,
        )
        parser.add_argument(
            "-n",
            "--num-threads",
            help="Set the number of threads for peak calculation",
            default=None,
            type=int,
        )
    else:
        parser.add_argument(
            "-ai",
            "--arithmetic-intensity",
            type=str,
            help="AI intensity input",
        )
        parser.add_argument(
            "-op", "--operations", type=str, help="Operations input"
        )

    return parser.parse_args()


def plot(args):

    try:
        fai = open(args.arithmetic_intensity, "r")
        fop = open(args.operations, "r")

        ai_data = json.load(fai)
        op_data = json.load(fop)

        _ai_ranks = []
        _op_ranks = []
        _data = {}

        for key, data in ai_data["timemory"].items():
            _ai_ranks.append(int(data["num_ranks"]))
            _data[key] = [data]

        for key, data in op_data["timemory"].items():
            _op_ranks.append(int(data["num_ranks"]))
            if key not in _data.keys():
                raise RuntimeError(
                    "Key '{}' found in operation data but not in AI data".format(
                        key
                    )
                )
            _data[key].append(data)

        for rank_ai, rank_op in zip(_ai_ranks, _op_ranks):
            if rank_ai != rank_op:
                raise RuntimeError(
                    "Number of ranks in output files is different: {} vs. {}".format(
                        len(rank_ai), len(rank_op)
                    )
                )

        use_label = True if len(_data) > 1 else False
        use_rank = True if len(_ai_ranks) > 1 else False

        _rank = 0
        for key, data in _data.items():
            plot_impl(
                args,
                data[0],
                data[1],
                _rank if use_rank else None,
                key if use_label else None,
            )
            _rank += 1

    except Exception as e:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback, limit=20)
        print("Exception - {}".format(e))
        sys.exit(1)

    if _errc != 0:
        print(f"Done (with errors) - {sys.argv[0]} (exit code: {_errc})")
    else:
        print(f"Done - {sys.argv[0]}")

    sys.exit(_errc)


def plot_impl(args, ai_data, op_data, rank=None, label=None):

    fname = os.path.basename(args.output_file)
    fdir = os.path.realpath(args.output_dir)

    if label is not None:
        fname = "_".join([label, fname])

    band_labels = [element.upper() for element in args.bandwidths]

    if rank is not None:
        fname = "{}_{}".format(fname, rank)
        args.title = "{} (MPI rank: {})".format(args.title, rank)

    import timemory.roofline as _roofline

    _roofline.plot_roofline(
        ai_data,
        op_data,
        band_labels,
        args.txns_bandwidths,
        args.inst_peak,
        args.rtype,
        args.display,
        fname,
        args.format,
        fdir,
        args.title,
        args.plot_dimensions[0],
        args.plot_dimensions[1],
        args.plot_dimensions[2],
        args.echo_dart,
        rank,
    )


def run(args, cmd):
    global _errc

    if len(cmd) == 0:
        return

    os.environ["TIMEMORY_PYTHON_BINDINGS"] = "0"

    def get_environ(env_var, default_value, dtype):
        val = os.environ.get(env_var)
        if val is None:
            os.environ[env_var] = "{}".format(default_value)
            return dtype(default_value)
        else:
            return dtype(val)

    _base_output_path = os.path.basename(cmd[0]).replace("_", "-")
    _default_output_path = "-".join(
        ["timemory", _base_output_path, "output"]
    ).replace("--", "-")
    output_path = get_environ("TIMEMORY_OUTPUT_PATH", _default_output_path, str)
    output_prefix = get_environ("TIMEMORY_OUTPUT_PREFIX", "", str)
    os.environ["TIMEMORY_JSON_OUTPUT"] = "ON"

    if args.num_threads is not None:
        os.environ["TIMEMORY_ROOFLINE_NUM_THREADS"] = "{}".format(
            args.num_threads
        )

    if args.preload:
        # walk back up the tree until we find libtimemory-preload.<EXT>
        preload = os.path.realpath(os.path.dirname(__file__))
        libname = None
        preload_env = None

        import platform

        if platform.system() == "Linux":
            libname = "libtimemory.so"
            preload_env = "LD_PRELOAD"
        elif platform.system() == "Darwin":
            libname = "libtimemory.dylib"
            preload_env = "DYLD_INSERT_LIBRARIES"
            os.environ["DYLD_FORCE_FLAT_NAMESPACE"] = "1"

        # platform support pre-loading
        if libname is not None:
            for i in range(0, 5):
                if os.path.exists(os.path.join(preload, libname)):
                    preload = os.path.join(preload, libname)
                    break
                else:
                    preload = os.path.dirname(preload)

        if preload_env is not None:
            current_preload = os.environ.get(preload_env)
            if current_preload is not None:
                os.environ[preload_env] = "{}:{}".format(
                    current_preload, preload
                )
            else:
                os.environ[preload_env] = "{}".format(preload)
        elif libname is not None:
            print(
                "Warning! Unable to locate '{}'. Preloading failed".format(
                    libname
                )
            )

    def handle_error(ret, cmd, keep_going):
        global _errc

        err_msg = "Error executing: '{}'".format(" ".join(cmd))
        if ret != 0 and not keep_going:
            raise RuntimeError(err_msg)
        elif ret != 0 and keep_going:
            barrier = "=" * 80
            err_msg = (
                "\n\n"
                + barrier
                + "\n\n    ERROR: "
                + err_msg
                + "\n\n"
                + barrier
                + "\n\n"
            )
            sys.stderr.write(err_msg)
            sys.stderr.flush()
            _errc = ret

    import subprocess as sp

    if args.rerun is None or args.rerun == "ai":
        os.environ["TIMEMORY_ROOFLINE_MODE"] = "ai"
        os.environ["TIMEMORY_ROOFLINE_MODE_CPU"] = "ai"
        os.environ["TIMEMORY_ROOFLINE_MODE_GPU"] = "ai"
        p = sp.Popen(cmd)
        ret = p.wait()
        handle_error(ret, cmd, args.keep_going)

    if args.rerun is None or args.rerun == "op":
        os.environ["TIMEMORY_ROOFLINE_MODE"] = "op"
        os.environ["TIMEMORY_ROOFLINE_MODE_CPU"] = "op"
        os.environ["TIMEMORY_ROOFLINE_MODE_GPU"] = "op"
        p = sp.Popen(cmd)
        ret = p.wait()
        handle_error(ret, cmd, args.keep_going)

    if "gpu_roofline" in args.rtype:
        args.arithmetic_intensity = os.path.join(
            output_path,
            "{}{}_activity.json".format(output_prefix, args.rtype),
        )
        args.operations = os.path.join(
            output_path,
            "{}{}_counters.json".format(output_prefix, args.rtype),
        )
    else:
        args.arithmetic_intensity = os.path.join(
            output_path, "{}{}_ai.json".format(output_prefix, args.rtype)
        )
        args.operations = os.path.join(
            output_path, "{}{}_op.json".format(output_prefix, args.rtype)
        )


def try_plot():
    global _errc

    try:
        # look for "--" and interpret anything after that
        # to be a command to execute
        _argv = []
        _cmd = []

        _argsets = [_argv, _cmd]
        _i = 0
        _separator = "--"

        for _arg in sys.argv[1:]:
            if _arg == _separator and _i < len(_argsets):
                _i += 1
            else:
                _argsets[_i].append(_arg)

        sys.argv[1:] = _argv

        args = parse_args(len(_cmd) != 0)
        run(args, _cmd)
        if args.verbose is not None:
            import timemory.roofline as _roofline

            _roofline.VERBOSE = args.verbose
        plot(args)

    except Exception as e:
        print(f"\nException :: command line argument error\n\t{e}\n")
        # warnings.warn(msg)
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback, limit=20)
        _errc = 1

    sys.exit(_errc)


if __name__ == "__main__":
    try_plot()
    sys.exit(_errc)



# JSON output test cases

import json
import unittest

import calipertest as cat

class CaliperJSONTest(unittest.TestCase):
    """ Caliper JSON formatters test cases """

    def test_jsonobject(self):
        """ Test json object layout """

        target_cmd = [ './ci_test_macros' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query',
                       '-q', 'SELECT count(),* group by prop:nested where function format json(object)' ]

        caliper_config = {
            'CALI_CONFIG_PROFILE'    : 'serial-trace',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        obj = json.loads( cat.run_test_with_query(target_cmd, query_cmd, caliper_config ) )

        self.assertTrue( { 'records', 'globals', 'attributes' }.issubset(set(obj.keys())) )

        self.assertEqual(len(obj['records']), 7)
        self.assertTrue('path' in obj['records'][0].keys())

        self.assertTrue('cali.caliper.version' in obj['globals'].keys())
        self.assertTrue('count' in obj['attributes'].keys())
        self.assertEqual(obj['attributes']['cali.caliper.version']['is_global'], 1)        

        
    def test_jsonobject_pretty(self):
        """ Test json object layout """

        target_cmd = [ './ci_test_macros' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query',
                       '-q', 'SELECT count(),* group by prop:nested where function format json(object,pretty)' ]

        caliper_config = {
            'CALI_CONFIG_PROFILE'    : 'serial-trace',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        obj = json.loads( cat.run_test_with_query(target_cmd, query_cmd, caliper_config ) )

        self.assertTrue( { 'records', 'globals', 'attributes' }.issubset(set(obj.keys())) )

        self.assertEqual(len(obj['records']), 7)
        self.assertTrue('path' in obj['records'][0].keys())

        self.assertTrue('cali.caliper.version' in obj['globals'].keys())
        self.assertTrue('count' in obj['attributes'].keys())
        self.assertEqual(obj['attributes']['cali.caliper.version']['is_global'], True)

        
    def test_jsontree(self):
        """ Test basic json-tree formatter """

        target_cmd = [ './ci_test_macros' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query',
                       '-q', 'SELECT count(),sum(time.inclusive.duration),loop,iteration#mainloop group by loop,iteration#mainloop format json-split' ]

        caliper_config = {
            'CALI_CONFIG_PROFILE'    : 'serial-trace',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        obj = json.loads( cat.run_test_with_query(target_cmd, query_cmd, caliper_config) )

        self.assertTrue( { 'data', 'columns', 'column_metadata', 'nodes' }.issubset(set(obj.keys())) )

        columns = obj['columns']

        self.assertEqual( { 'path', 'iteration#mainloop', 'count', 'sum#time.inclusive.duration' }, set(columns) )

        data = obj['data']

        self.assertEqual(len(data), 10)
        self.assertEqual(len(data[0]), 4)

        meta = obj['column_metadata']

        self.assertEqual(len(meta), 4)
        self.assertTrue(meta[columns.index('count')]['is_value'])
        self.assertFalse(meta[columns.index('path')]['is_value'])

        nodes = obj['nodes']

        self.assertEqual(nodes[0]['label' ], 'mainloop')
        self.assertEqual(nodes[0]['column'], 'path')
        self.assertEqual(nodes[1]['label' ], 'fooloop')
        self.assertEqual(nodes[1]['column'], 'path')
        self.assertEqual(nodes[1]['parent'], 0)

        iterindex = columns.index('iteration#mainloop')

        # Note: this is a pretty fragile test
        self.assertEqual(data[9][iterindex], 3)

        
    def test_hatchetcontroller(self):
        """ Test hatchet-region-profile controller """

        target_cmd = [ './ci_test_macros', '0', 'hatchet-region-profile,use.mpi=false,output=stdout' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'     : '0'
        }

        obj = json.loads( cat.run_test(target_cmd, caliper_config)[0] )

        self.assertTrue( { 'data', 'columns', 'column_metadata', 'nodes' }.issubset(set(obj.keys())) )

        columns = obj['columns']

        self.assertEqual( { 'path', 'time' }, set(columns) )

        data = obj['data']

        self.assertEqual(len(data), 8)
        self.assertEqual(len(data[0]), 2)

        meta = obj['column_metadata']

        self.assertEqual(len(meta), 2)

        time_idx = columns.index('time')
        path_idx = columns.index('path')

        self.assertTrue( meta[time_idx]['is_value'])
        self.assertEqual(meta[time_idx]['attribute.unit'], 'sec')
        self.assertFalse(meta[path_idx]['is_value'])


    def test_esc(self):
        target_cmd = [ './ci_test_basic' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query',
                       '-q', 'select *,count() format json-split' ]

        caliper_config = {
            'CALI_CONFIG_PROFILE'    : 'serial-trace',
            'CALI_RECORDER_FILENAME' : 'stdout',
            'CALI_LOG_VERBOSITY'     : '0'
        }

        obj = json.loads( cat.run_test_with_query(target_cmd, query_cmd, caliper_config) )

        columns = obj['columns']

        self.assertTrue('event.set# =\\weird ""attribute"=  ' in columns)

        data  = obj['data']
        nodes = obj['nodes']

        index = columns.index('event.set# =\\weird ""attribute"=  ')

        self.assertEqual(nodes[data[0][index]]['label'], '  \\\\ weird," name",' )
        self.assertEqual(obj[' =\\weird "" global attribute"=  '], '  \\\\ weird," name",')

        
if __name__ == "__main__":
    unittest.main()



# Sampler/symbollookup tests

import unittest

import calipertest as cat

class CaliperSamplerTest(unittest.TestCase):
    """ Caliper sampler test case """

    def test_hatchet_sample_profile(self):
        target_cmd = [ './ci_test_macros', '5000', 'hatchet-sample-profile(use.mpi=false,output=stdout,output.format=cali)' ]
        query_cmd  = [ '../../src/tools/cali-query/cali-query', '-e' ]

        caliper_config = {
            'CALI_LOG_VERBOSITY'     : '0'
        }

        query_output = cat.run_test_with_query(target_cmd, query_cmd, caliper_config)
        snapshots = cat.get_snapshots_from_text(query_output)

        self.assertTrue(len(snapshots) > 0)

        self.assertTrue(cat.has_snapshot_with_keys(
            snapshots, { 'loop', 'function' }))

if __name__ == "__main__":
    unittest.main()



import unittest

import calipertest as cat


class CaliperValidatorTest(unittest.TestCase):
    """ Caliper validator service test cases """

    def test_validator_nesting_error(self):
        target_cmd = [ '../cali-test', 'nesting-error' ]

        env = {
            'CALI_LOG_VERBOSITY'   : '1',
            'CALI_LOG_LOGFILE'     : 'stdout',
            'CALI_SERVICES_ENABLE' : 'validator'
        }

        log_targets = [
            'validator: incorrect nesting: trying to end "test.nesting-error.a"="11" but current attribute is "test.nesting-error.b"',
            'validator: Annotation nesting errors found'
        ]

        report_out,_ = cat.run_test(target_cmd, env)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

    def test_validator_unclosed_region(self):
        target_cmd = [ '../cali-test', 'unclosed-region' ]

        env = {
            'CALI_LOG_VERBOSITY'   : '1',
            'CALI_LOG_LOGFILE'     : 'stdout',
            'CALI_SERVICES_ENABLE' : 'validator'
        }

        log_targets = [
            'validator: Regions not closed: test.unclosed_region=101',
            'validator: Annotation nesting errors found'
        ]

        report_out,_ = cat.run_test(target_cmd, env)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

    def test_validator_pass_blob(self):
        target_cmd = [ '../cali-test', 'blob' ]

        env = {
            'CALI_LOG_VERBOSITY'   : '1',
            'CALI_LOG_LOGFILE'     : 'stdout',
            'CALI_SERVICES_ENABLE' : 'validator'
        }

        log_targets = [
            'validator: No annotation nesting errors found'
        ]

        report_out,_ = cat.run_test(target_cmd, env)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)

    def test_validator_pass_macros(self):
        target_cmd = [ './ci_test_macros', '0' ]

        env = {
            'CALI_LOG_VERBOSITY'   : '1',
            'CALI_LOG_LOGFILE'     : 'stdout',
            'CALI_SERVICES_ENABLE' : 'validator'
        }

        log_targets = [
            'validator: No annotation nesting errors found'
        ]

        report_out,_ = cat.run_test(target_cmd, env)
        lines = report_out.decode().splitlines()

        for target in log_targets:
            for line in lines:
                if target in line:
                    break
            else:
                self.fail('%s not found in log' % target)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import sys
import argparse
# read in components and mangled enums
from timemory_types import native_components


def generate_extern(component, key, alias, suffix, specify_comp=True):
    """
    This function generates a type trait label for C++
    """
    if specify_comp:
        return "TIMEMORY_{}_EXTERN_{}({}, {})".format(
            key.upper(), suffix.upper(), alias, "::tim::component::{}".format(component))
    else:
        return "TIMEMORY_{}_EXTERN_{}({})".format(
            key.upper(), suffix.upper(), alias)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-V", "--verbose",
                        help="Enable verbosity", default=0, type=int)
    parser.add_argument(
        "-i", "--indent", help="Indentation", default=4, type=int)
    args = parser.parse_args()

    if args.verbose > 0:
        print("timemory components: [{}]\n".format(
            ", ".join(native_components)))

    native_components.sort()
    outdata = "\n"
    for key in ["declare", "instantiate"]:
        for suffix in ["tuple", "list", "hybrid"]:
            specify_comp = False if suffix == "hybrid" else True
            for component in native_components:
                outdata += "{}\n".format(generate_extern(
                    component, key, "_".join([component, "t"]),
                    suffix, specify_comp))
            outdata += "\n"

    print("{}".format(outdata))



#!@PYTHON_EXECUTABLE@
#
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import threading
from functools import wraps

from ..libs.libpytimemory.trace import tracer_function as _tracer_function
from ..libs.libpytimemory.trace import config as _tracer_config
from ..libs.libpytimemory.trace import tracer_init as _tracer_init
from ..libs.libpytimemory.trace import tracer_finalize as _tracer_fini
from ..libs.libpytimemory.trace import trace_bundle as _tracer_bundle
from ..libs.libpytimemory import settings


__all__ = ["trace", "config", "Tracer", "FakeTracer", "Config"]


#
def _default_functor():
    return True


#
PY3 = sys.version_info[0] == 3
PY35 = PY3 and sys.version_info[1] >= 5

# exec (from https://bitbucket.org/gutworth/six/):
if PY3:
    import builtins

    exec_ = getattr(builtins, "exec")
    del builtins
else:

    def exec_(_code_, _globs_=None, _locs_=None):
        """Execute code in a namespace."""
        if _globs_ is None:
            frame = sys._getframe(1)
            _globs_ = frame.f_globals
            if _locs_ is None:
                _locs_ = frame.f_locals
            del frame
        elif _locs_ is None:
            _locs_ = _globs_
        exec("""exec _code_ in _globs_, _locs_""")


config = _tracer_config
Config = _tracer_config


class Tracer:
    """Provides decorators and context-manager for the timemory tracers"""

    global _default_functor

    # static variable
    _conditional_functor = _default_functor

    # ---------------------------------------------------------------------------------- #
    #
    @staticmethod
    def condition(functor):
        """Assign a function evaluating whether to enable the tracer"""
        Tracer._conditional_functor = functor

    # ---------------------------------------------------------------------------------- #
    #
    @staticmethod
    def is_enabled():
        """Checks whether the tracer is enabled"""

        try:
            return Tracer._conditional_functor()
        except Exception:
            pass
        return False

    # ---------------------------------------------------------------------------------- #
    #
    def __init__(self, components=[], **kwargs):
        """
        Arguments:
            - components [list, str, or function]  : list of timemory components
        """

        self.components = components
        self._original_function = sys.gettrace()
        self._unset = 0
        self._use = (
            not _tracer_config._is_running and Tracer.is_enabled() is True
        )
        self.debug = kwargs["debug"] if "debug" in kwargs else False

    # ---------------------------------------------------------------------------------- #
    #
    def __del__(self):
        """Make sure the tracer stops"""

        self.stop()

    # ---------------------------------------------------------------------------------- #
    #
    def configure(self):
        """Initialize, configure the bundle, store original tracer function"""

        _tracer_init()
        _trace = settings.trace_components
        _profl = settings.profiler_components
        _components = _profl if _trace is None else _trace
        if len(_components) == 0:
            _components = settings.global_components

        # support function and list
        import inspect

        if inspect.isfunction(self.components):
            self.components = self.components()

        input_components = []
        if isinstance(self.components, list):
            input_components = self.components
        else:
            input_components = ["{}".format(self.components)]

        # factor in global and local settings
        components = input_components + _components.split(",")
        components = list(set([v.lower() for v in components]))
        components = [f"{v}" for v in components]
        if "" in components:
            components.remove("")

        # update global setting
        if _trace is None or len(_trace) == 0:
            settings.trace_components = ",".join(components)

        # configure
        if settings.debug or self.debug:
            sys.stderr.write(
                "configuring Profiler with components (type: {}): {}\n".format(
                    type(components).__name__, components
                )
            )
        try:
            _tracer_bundle.configure(components, True, False)
        except Exception as e:
            sys.stderr.write(f"Tracer configuration failed: {e}\n")

        # store original
        if settings.debug or self.debug:
            sys.stderr.write("setting trace function...\n")
        self._original_function = sys.gettrace()

        if settings.debug or self.debug:
            sys.stderr.write("Tracer configured...\n")

    # ---------------------------------------------------------------------------------- #
    #
    def update(self):
        """Updates whether the profiler is already running based on whether the tracer
        is not already running, is enabled, and the function is not already set
        """

        self._use = (
            not _tracer_config._is_running
            and Tracer.is_enabled() is True
            and sys.gettrace() == self._original_function
        )

    # ---------------------------------------------------------------------------------- #
    #
    def start(self):
        """Start the tracer"""

        self.update()
        if self._use:
            if settings.debug or self.debug:
                sys.stderr.write("Tracer starting...\n")
            self.configure()
            sys.settrace(_tracer_function)
            threading.settrace(_tracer_function)
            if settings.debug or self.debug:
                sys.stderr.write("Tracer started...\n")

        self._unset = self._unset + 1
        return self._unset

    # ---------------------------------------------------------------------------------- #
    #
    def stop(self):
        """Stop the tracer"""

        self._unset = self._unset - 1
        if self._unset == 0:
            if settings.debug or self.debug:
                sys.stderr.write("Tracer stopping...\n")
            sys.settrace(self._original_function)
            _tracer_fini()
            if settings.debug or self.debug:
                sys.stderr.write("Tracer stopped...\n")

        return self._unset

    # ---------------------------------------------------------------------------------- #
    #
    def __call__(self, func):
        """Decorator"""

        @wraps(func)
        def function_wrapper(*args, **kwargs):
            # store whether this tracer started
            self.start()
            # execute the wrapped function
            result = func(*args, **kwargs)
            # unset the tracer if this wrapper set it
            self.stop()
            # return the result of the wrapped function
            return result

        return function_wrapper

    # ---------------------------------------------------------------------------------- #
    #
    def __enter__(self, *args, **kwargs):
        """Context manager start function"""

        self.start()

    # ---------------------------------------------------------------------------------- #
    #
    def __exit__(self, exec_type, exec_value, exec_tb):
        """Context manager stop function"""

        self.stop()

        if (
            exec_type is not None
            and exec_value is not None
            and exec_tb is not None
        ):
            import traceback

            traceback.print_exception(exec_type, exec_value, exec_tb, limit=5)

    # ---------------------------------------------------------------------------------- #
    #
    def run(self, cmd):
        """Execute and trace a command"""

        import __main__

        dict = __main__.__dict__
        if isinstance(cmd, str):
            return self.runctx(cmd, dict, dict)
        else:
            return self.runctx(" ".join(cmd), dict, dict)

    # ---------------------------------------------------------------------------------- #
    #
    def runctx(self, cmd, _globals, _locals):
        """Trace a context"""

        try:
            self.start()
            exec_(cmd, _globals, _locals)
        finally:
            self.stop()

        return self

    # ---------------------------------------------------------------------------------- #
    #
    def runcall(self, func, *args, **kw):
        """Trace a single function call"""

        try:
            self.start()
            return func(*args, **kw)
        finally:
            self.stop()


trace = Tracer


class FakeTracer:
    """Provides dummy decorators and context-manager for the timemory tracer"""

    # ---------------------------------------------------------------------------------- #
    #
    @staticmethod
    def condition(functor):
        pass

    # ---------------------------------------------------------------------------------- #
    #
    @staticmethod
    def is_enabled():
        return False

    # ---------------------------------------------------------------------------------- #
    #
    def __init__(self, *args, **kwargs):
        """
        Arguments:
            - components [list of strings]  : list of timemory components
            - flat [bool]                   : enable flat profiling
            - timeline [bool]               : enable timeline profiling
        """
        pass

    # ---------------------------------------------------------------------------------- #
    #
    def __call__(self, func):
        """Decorator"""

        @wraps(func)
        def function_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return function_wrapper

    # ---------------------------------------------------------------------------------- #
    #
    def __enter__(self, *args, **kwargs):
        """Context manager begin"""
        pass

    # ---------------------------------------------------------------------------------- #
    #
    def __exit__(self, exec_type, exec_value, exec_tb):
        """Context manager end"""

        import traceback

        if (
            exec_type is not None
            and exec_value is not None
            and exec_tb is not None
        ):
            traceback.print_exception(exec_type, exec_value, exec_tb, limit=5)



#!@PYTHON_EXECUTABLE@
#
# MIT License
#
# Copyright (c) 2018, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S. Dept. of Energy).  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Imports timemory.libpytimemory.options as timemory.options
"""

from __future__ import absolute_import

__author__ = "Jonathan Madsen"
__copyright__ = "Copyright 2020, The Regents of the University of California"
__credits__ = ["Jonathan Madsen"]
__license__ = "MIT"
__version__ = "@PROJECT_VERSION@"
__maintainer__ = "Jonathan Madsen"
__email__ = "jrmadsen@lbl.gov"
__status__ = "Development"

from .libs.libpytimemory.options import *  # noqa: F401



# Copyright (c) 2020, Lawrence Livermore National Security, LLC.
# See top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

from .caliperreader import CaliperReader, read_caliper_globals, read_caliper_contents
from .caliperstreamreader import CaliperStreamReader



# Copyright 2017-2021 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from hatchet import GraphFrame


def test_graphframe(mock_graph_literal):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_literal(mock_graph_literal)

    assert len(gf.dataframe) == 24


def test_graphframe_to_literal(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    graph_literal = gf.to_literal()

    test_literal_output = gf.from_literal(graph_literal)

    assert len(test_literal_output.graph.roots) == len(gf.graph.roots)
    assert len(test_literal_output.graph) == len(gf.graph)


def test_with_duplicates(mock_graph_literal_duplicates):
    gf = GraphFrame.from_literal(mock_graph_literal_duplicates)

    assert len(gf.graph) == 6

    graph_literal = gf.to_literal()
    assert mock_graph_literal_duplicates.sort() == graph_literal.sort()


def test_with_duplicate_in_first_node(mock_graph_literal_duplicate_first):
    gf = GraphFrame.from_literal(mock_graph_literal_duplicate_first)

    assert len(gf.graph) == 6

    graph_literal = gf.to_literal()
    assert mock_graph_literal_duplicate_first.sort() == graph_literal.sort()



#!@PYTHON_EXECUTABLE@

import timemory
from timemory.bundle import auto_timer
import sys


def fib(n):
    return n if n < 2 else (fib(n - 1) + fib(n - 2))


@auto_timer("total", mode="full")
def main(nfib):

    ans = fib(nfib)
    with auto_timer("nested"):
        ans += fib(nfib + 1)

    print("Answer = {}".format(ans))


if __name__ == "__main__":
    main(int(sys.argv[1]) if len(sys.argv) > 1 else 34)
    timemory.timemory_finalize()



# Copyright 2017-2021 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from functools import total_ordering

from .frame import Frame


def traversal_order(node):
    """Deterministic key function for sorting nodes in traversals."""
    return (node.frame, id(node))


@total_ordering
class Node:
    """A node in the graph. The node only stores its frame."""

    def __init__(self, frame_obj, parent=None, hnid=-1, depth=-1):
        self.frame = frame_obj
        self._depth = depth
        self._hatchet_nid = hnid

        self.parents = []
        if parent is not None:
            self.add_parent(parent)
        self.children = []

    def add_parent(self, node):
        """Adds a parent to this node's list of parents."""
        assert isinstance(node, Node)
        self.parents.append(node)

    def add_child(self, node):
        """Adds a child to this node's list of children."""
        assert isinstance(node, Node)
        self.children.append(node)

    def paths(self, attrs=None):
        """List of tuples, one for each path from this node to any root.

        Arguments:
            attrs (str or list, optional): attribute(s) to extract from Frames

        Paths are tuples of Frame objects, or, if attrs is provided, they
        are paths containing the requested attributes.
        """
        node_value = (self.frame,) if attrs is None else (self.frame.values(attrs),)
        if not self.parents:
            return [node_value]
        else:
            paths = []
            for parent in self.parents:
                parent_paths = parent.paths(attrs)
                paths.extend([path + node_value for path in parent_paths])
            return paths

    def path(self, attrs=None):
        """Path to this node from root. Raises if there are multiple paths.

        Arguments:
            attrs (str or list, optional): attribute(s) to extract from Frames

        This is useful for trees (where each node only has one path), as
        it just gets the only element from ``self.paths``.  This will
        fail with a MultiplePathError if there is more than one path to
        this node.
        """
        paths = self.paths(attrs)
        if len(paths) > 1:
            raise MultiplePathError("Node has more than one path: " % paths)
        return paths[0]

    def dag_equal(self, other, vs=None, vo=None):
        """Check if DAG rooted at self has the same structure as that rooted at
        other.
        """
        if vs is None:
            vs = set()
        if vo is None:
            vo = set()

        vs.add(self._hatchet_nid)
        vo.add(other._hatchet_nid)

        # if number of children do not match, then nodes are not equal
        if len(self.children) != len(other.children):
            return False

        # sort children of each node by its frame
        ssorted = sorted(self.children, key=lambda x: x.frame)
        osorted = sorted(other.children, key=lambda x: x.frame)

        for self_child, other_child in zip(ssorted, osorted):
            # if frames do not match, then nodes are not equal
            if self_child.frame != other_child.frame:
                return False

            visited_s = self_child._hatchet_nid in vs
            visited_o = other_child._hatchet_nid in vo

            # check for duplicate nodes
            if visited_s != visited_o:
                return False

            # skip visited nodes
            if visited_s or visited_o:
                continue

            # recursive check for node equality
            if not self_child.dag_equal(other_child, vs, vo):
                return False

        return True

    def traverse(self, order="pre", attrs=None, visited=None):
        """Traverse the tree depth-first and yield each node.

        Arguments:
            order (str):  "pre" or "post" for preorder or postorder (default: pre)
            attrs (list or str, optional): if provided, extract these fields
                from nodes while traversing and yield them
            visited (dict, optional): dictionary in which each visited
                node's in-degree will be stored
        """
        if order not in ("pre", "post"):
            raise ValueError("order must be one of 'pre' or 'post'")

        if visited is None:
            visited = {}

        key = id(self)
        if key in visited:
            # count the number of times we reached
            visited[key] += 1
            return
        visited[key] = 1

        def value(node):
            return node if attrs is None else node.frame.values(attrs)

        if order == "pre":
            yield value(self)

        for child in sorted(self.children, key=traversal_order):
            for item in child.traverse(order=order, attrs=attrs, visited=visited):
                yield item

        if order == "post":
            yield value(self)

    def __hash__(self):
        return self._hatchet_nid

    def __eq__(self, other):
        return self._hatchet_nid == other._hatchet_nid

    def __lt__(self, other):
        return self._hatchet_nid < other._hatchet_nid

    def __gt__(self, other):
        return self._hatchet_nid > other._hatchet_nid

    def __str__(self):
        """Returns a string representation of the node."""
        return str(self.frame)

    def copy(self):
        """Copy this node without preserving parents or children."""
        return Node(frame_obj=self.frame.copy())

    @classmethod
    def from_lists(cls, lists):
        r"""Construct a hierarchy of nodes from recursive lists.

For example, this will construct a simple tree:

.. code-block:: python

    Node.from_lists(
        ["a",
            ["b", "d", "e"],
            ["c", "f", "g"],
        ]
    )

.. code-block:: console

         a
        / \
       b   c
     / |   | \
    d  e   f  g

And this will construct a simple diamond DAG:

.. code-block:: python

    d = Node(Frame(name="d"))
    Node.from_lists(
        ["a",
            ["b", d],
            ["c", d]
        ]
    )

.. code-block:: console

      a
     / \
    b   c
     \ /
      d

In the above examples, the 'a' represents a Node with its
`frame == Frame(name="a")`.
"""

        def _from_lists(lists, parent):
            if isinstance(lists, (tuple, list)):
                if isinstance(lists[0], Node):
                    node = lists[0]
                elif isinstance(lists[0], str):
                    node = Node(Frame(name=lists[0]))
                children = lists[1:]
                for val in children:
                    _ = _from_lists(val, node)
            elif isinstance(lists, str):
                node = Node(Frame(name=lists))
            elif isinstance(lists, Node):
                node = lists
            else:
                raise ValueError("Argument must be str, list, or Node: %s" % lists)

            if parent:
                node.add_parent(parent)
                parent.add_child(node)

            return node

        return _from_lists(lists, None)

    def __repr__(self):
        return "Node({%s})" % ", ".join(
            "%s: %s" % (repr(k), repr(v)) for k, v in sorted(self.frame.attrs.items())
        )


class MultiplePathError(Exception):
    """Raised when a node is asked for a single path but has multiple."""
